#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

cdef extern from "<complex>" namespace "std" nogil:
    double complex conj(double complex x)

from qutip.core.data.base cimport idxint, Data
from qutip.core.data cimport csr, dense
from qutip.core.data.csr cimport CSR
from qutip.core.data.dense cimport Dense
from qutip.core.data.dia cimport Dia
from qutip.core.data.matmul cimport matmul_dense
from .matmul import matmul
from .trace import trace
from .adjoint import adjoint

__all__ = [
    'inner', 'inner_csr', 'inner_dense', 'inner_dia', 'inner_data',
    'inner_op', 'inner_op_csr', 'inner_op_dense', 'inner_op_dia',
    'inner_op_data',
]


cdef int _check_shape_inner(Data left, Data right) except -1 nogil:
    if (
        (left.shape[0] != 1 and left.shape[1] != 1)
        or right.shape[1] != 1
    ):
        raise ValueError(
            "incompatible matrix shapes "
            + str(left.shape)
            + " and "
            + str(right.shape)
        )
    return 0

cdef int _check_shape_inner_op(Data left, Data op, Data right) except -1 nogil:
    cdef bint left_shape = left.shape[0] == 1 or left.shape[1] == 1
    cdef bint left_op = (
        (left.shape[0] == 1 and left.shape[1] == op.shape[0])
        or (left.shape[1] == 1 and left.shape[0] == op.shape[0])
    )
    cdef bint op_right = op.shape[1] == right.shape[0]
    cdef bint right_shape = right.shape[1] == 1
    if not (left_shape and left_op and op_right and right_shape):
        raise ValueError("".join([
            "incompatible matrix shapes ",
            str(left.shape),
            ", ",
            str(op.shape),
            " and ",
            str(right.shape),
        ]))
    return 0

cdef double complex _inner_csr_bra_ket(CSR left, CSR right) nogil:
    cdef size_t col, ptr_bra, ptr_ket
    cdef double complex out = 0
    # We actually don't care if left is sorted or not.
    for ptr_bra in range(csr.nnz(left)):
        col = left.col_index[ptr_bra]
        ptr_ket = right.row_index[col]
        if right.row_index[col + 1] != ptr_ket:
            out += left.data[ptr_bra] * right.data[ptr_ket]
    return out

cdef double complex _inner_csr_ket_ket(CSR left, CSR right) nogil:
    cdef size_t row, ptr_l, ptr_r
    cdef double complex out = 0
    for row in range(left.shape[0]):
        ptr_l = left.row_index[row]
        ptr_r = right.row_index[row]
        if left.row_index[row+1] != ptr_l and right.row_index[row+1] != ptr_r:
            out += conj(left.data[ptr_l]) * right.data[ptr_r]
    return out

cpdef double complex inner_csr(CSR left, CSR right, bint scalar_is_ket=False) except *:
    """
    Compute the complex inner product <left|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner(left, right)
    if left.shape[0] == left.shape[1] == right.shape[1] == 1:
        if csr.nnz(left) and csr.nnz(right):
            return (
                conj(left.data[0]) * right.data[0] if scalar_is_ket
                else left.data[0] * right.data[0]
            )
        return 0
    if left.shape[0] == 1:
        return _inner_csr_bra_ket(left, right)
    return _inner_csr_ket_ket(left, right)

cpdef double complex inner_dia(Dia left, Dia right, bint scalar_is_ket=False) except * nogil:
    """
    Compute the complex inner product <left|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner(left, right)
    cdef double complex inner = 0.
    cdef idxint diag_left, diag_right
    cdef bint is_ket
    if right.shape[0] == 1:
        is_ket = scalar_is_ket
    else:
        is_ket = left.shape[0] == right.shape[0]

    if is_ket:
      for diag_right in range(right.num_diag):
        for diag_left in range(left.num_diag):
          if left.offsets[diag_left] - right.offsets[diag_right] == 0:
            inner += (
              conj(left.data[diag_left * left.shape[1]])
              * right.data[diag_right * right.shape[1]]
            )
    else:
      for diag_right in range(right.num_diag):
        for diag_left in range(left.num_diag):
          if left.offsets[diag_left] + right.offsets[diag_right] == 0:
            inner += (
              left.data[diag_left * left.shape[1] + left.offsets[diag_left]]
              * right.data[diag_right * right.shape[1]]
            )

    return inner

cpdef double complex inner_dense(Dense left, Dense right, bint scalar_is_ket=False) except * nogil:
    """
    Compute the complex inner product <left|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner(left, right)
    if left.shape[0] == left.shape[1] == right.shape[1] == 1:
        return (
                conj(left.data[0]) * right.data[0] if scalar_is_ket
                else left.data[0] * right.data[0]
        )
    cdef double complex out = 0
    cdef size_t i
    if left.shape[0] == 1:
        for i in range(right.shape[0]):
            out += left.data[i] * right.data[i]
    else:
        for i in range(right.shape[0]):
            out += conj(left.data[i]) * right.data[i]
    return out


cdef double complex _inner_op_csr_bra_ket(CSR left, CSR op, CSR right) nogil:
    cdef size_t ptr_l, ptr_op, ptr_r, row, col
    cdef double complex sum, out=0
    # left does not need to be sorted.
    for ptr_l in range(csr.nnz(left)):
        row = left.col_index[ptr_l]
        sum = 0
        for ptr_op in range(op.row_index[row], op.row_index[row + 1]):
            col = op.col_index[ptr_op]
            ptr_r = right.row_index[col]
            if ptr_r != right.row_index[col + 1]:
                sum += op.data[ptr_op] * right.data[ptr_r]
        out += left.data[ptr_l] * sum
    return out

cdef double complex _inner_op_csr_ket_ket(CSR left, CSR op, CSR right) nogil:
    cdef size_t ptr_l, ptr_op, ptr_r, row, col
    cdef double complex sum, out=0
    for row in range(op.shape[0]):
        ptr_l = left.row_index[row]
        if left.row_index[row + 1] == ptr_l:
            continue
        sum = 0
        for ptr_op in range(op.row_index[row], op.row_index[row + 1]):
            col = op.col_index[ptr_op]
            ptr_r = right.row_index[col]
            if ptr_r != right.row_index[col + 1]:
                sum += op.data[ptr_op] * right.data[ptr_r]
        out += conj(left.data[ptr_l]) * sum
    return out

cpdef double complex inner_op_dia(Dia left, Dia op, Dia right,
                                   bint scalar_is_ket=False) except * nogil:
    """
    Compute the complex inner product <left|op|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner_op(left, op, right)
    cdef double complex inner = 0., val
    cdef idxint diag_left, diag_op, diag_right
    cdef int is_ket
    if op.shape[0] == 1:
        is_ket = scalar_is_ket
    else:
        is_ket = left.shape[0] == op.shape[0]

    if is_ket:
      for diag_right in range(right.num_diag):
        for diag_left in range(left.num_diag):
          for diag_op in range(op.num_diag):
            if -left.offsets[diag_left] + right.offsets[diag_right] + op.offsets[diag_op] == 0:
              inner += (
                conj(left.data[diag_left])
                * right.data[diag_right]
                * op.data[diag_op * op.shape[1] - right.offsets[diag_right]]
              )
    else:
      for diag_right in range(right.num_diag):
        for diag_left in range(left.num_diag):
          for diag_op in range(op.num_diag):
            if left.offsets[diag_left] + right.offsets[diag_right] + op.offsets[diag_op] == 0:
              inner += (
                left.data[diag_left * left.shape[1] + left.offsets[diag_left]]
                * right.data[diag_right]
                * op.data[diag_op * op.shape[1] - right.offsets[diag_right]]
              )

    return inner

cpdef double complex inner_op_csr(CSR left, CSR op, CSR right,
                                  bint scalar_is_ket=False) except *:
    """
    Compute the complex inner product <left|op|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner_op(left, op, right)
    cdef double complex l
    if 1 == left.shape[1] == left.shape[0] == op.shape[0] == op.shape[1] == right.shape[1]:
        if not (csr.nnz(left) and csr.nnz(op) and csr.nnz(right)):
            return 0
        l = conj(left.data[0]) if scalar_is_ket else left.data[0]
        return l * op.data[0] * right.data[0]
    if left.shape[0] == 1:
        return _inner_op_csr_bra_ket(left, op, right)
    return _inner_op_csr_ket_ket(left, op, right)

cpdef double complex inner_op_dense(Dense left, Dense op, Dense right,
                                  bint scalar_is_ket=False) except *:
    """
    Compute the complex inner product <left|op|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner_op(left, op, right)
    return inner_dense(left, matmul_dense(op, right), scalar_is_ket)


cpdef inner_data(Data left, Data right, bint scalar_is_ket=False):
    """
    Compute the complex inner product <left|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner(left, right)
    if left.shape[0] == left.shape[1] == right.shape[1] == 1:
        return (
            trace(left).conjugate() * trace(right) if scalar_is_ket
            else trace(left) * trace(right)
        )

    if left.shape[0] != 1:
        left = adjoint(left)
    # We use trace so we don't force convertion to complex.
    return trace(matmul(left, right))


cpdef inner_op_data(Data left, Data op, Data right, bint scalar_is_ket=False):
    """
    Compute the complex inner product <left|op|right>.  The shape of `left` is
    used to determine if it has been supplied as a ket or a bra.  The result of
    this function will be identical if passed `left` or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.
    """
    _check_shape_inner_op(left, op, right)
    return inner_data(left, matmul(op, right), scalar_is_ket)


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

inner = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('left', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('right', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('scalar_is_ket', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=False),
    ]),
    name='inner',
    module=__name__,
    inputs=('left', 'right'),
    out=False,
)
inner.__doc__ =\
    """
    Compute the complex inner product <left|right>.  Return the complex value.

    The shape of `left` is used to determine if it has been supplied as a ket
    or a bra.  The result of this function will be identical if passed `left`
    or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.

    Parameters
    ----------
    left : Data
        The left operand as either a bra or a ket matrix.

    right : Data
        The right operand as a ket matrix.

    scalar_is_ket : bool, optional (False)
        If `False`, then `left` is assumed to be a bra if it is
        one-dimensional.  If `True`, then it is assumed to be a ket.  This
        parameter is ignored if `left` and `right` are not one-dimensional.
    """
inner.add_specialisations([
    (CSR, CSR, inner_csr),
    (Dia, Dia, inner_dia),
    (Dense, Dense, inner_dense),
    (Data, Data, inner_data),
], _defer=True)

inner_op = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('left', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('op', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('right', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('scalar_is_ket', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=False),
    ]),
    name='inner_op',
    module=__name__,
    inputs=('left', 'op', 'right'),
    out=False,
)
inner_op.__doc__ =\
    """
    Compute the complex inner product <left|op|right>.  Return the complex
    value.  This operation is also known as taking a matrix element.

    The shape of `left` is used to determine if it has been supplied as a ket
    or a bra.  The result of this function will be identical if passed `left`
    or `adjoint(left)`.

    The parameter `scalar_is_ket` is only intended for the case where `left`
    and `right` are both of shape (1, 1).  In this case, `left` will be assumed
    to be a ket unless `scalar_is_ket` is False.  This parameter is ignored at
    all other times.

    Parameters
    ----------
    left : Data
        The left operand as either a bra or a ket matrix.

    op : Data
        The operator of which to take the matrix element.  Must have dimensions
        which match `left` and `right`.

    right : Data
        The right operand as a ket matrix.

    scalar_is_ket : bool, optional (False)
        If `False`, then `left` is assumed to be a bra if it is
        one-dimensional.  If `True`, then it is assumed to be a ket.  This
        parameter is ignored if `left` and `right` are not one-dimensional.
    """
inner_op.add_specialisations([
    (CSR, CSR, CSR, inner_op_csr),
    (Dia, Dia, Dia, inner_op_dia),
    (Dense, Dense, Dense, inner_op_dense),
    (Data, Data, Data, inner_op_data),
], _defer=True)

del _inspect, _Dispatcher



#cython: language_level=3

from qutip.core.data.csr cimport CSR
from qutip.core.data.dense cimport Dense
from qutip.core.data.dia cimport Dia

cpdef CSR adjoint_csr(CSR matrix)
cpdef CSR transpose_csr(CSR matrix)
cpdef CSR conj_csr(CSR matrix)

cpdef Dense adjoint_dense(Dense matrix)
cpdef Dense transpose_dense(Dense matrix)
cpdef Dense conj_dense(Dense matrix)

cpdef Dia adjoint_dia(Dia matrix)
cpdef Dia transpose_dia(Dia matrix)
cpdef Dia conj_dia(Dia matrix)



#
# This code was contributed by Ben Criger. Resemblance to
# partial_transpose is intentional, and meant to enhance legibility.
#

__all__ = ['subsystem_apply']

import itertools

import numpy as np

from . import Qobj, qeye, to_kraus, tensor
from . import data as _data


def subsystem_apply(state, channel, mask, reference=False):
    """
    Returns the result of applying the propagator `channel` to the
    subsystems indicated in `mask`, which comprise the density operator
    `state`.

    Parameters
    ----------
    state : :class:`.Qobj`
        A density matrix or ket.

    channel : :class:`.Qobj`
        A propagator, either an `oper` or `super`.

    mask : *list* / *array*
        A mask that selects which subsystems should be subjected to the
        channel.

    reference : bool
        Decides whether explicit Kraus map should be used to evaluate action
        of channel.

    Returns
    -------
    rho_out: :class:`.Qobj`
        A density matrix with the selected subsystems transformed
        according to the specified channel.
    """
    # TODO: Include sparse/dense methods a la partial_transpose.
    if not (state.isket or state.isoper):
        raise ValueError("input state must be a ket or oper")
    if not (channel.issuper or channel.isoper):
        raise ValueError("input channel must be a super or oper")
    # Since there's only one channel, all affected subsystems must have
    # the same dimensions:
    aff_subs_dim_ar = np.transpose(np.array(state.dims))[np.array(mask)]
    if any((aff_subs_dim_ar[j] != aff_subs_dim_ar[0]).any()
           for j in range(len(aff_subs_dim_ar))):
        raise ValueError("affected subsystems must have the same dimension")
    # If the map is on the Hilbert space, it must have the same dimension
    # as the affected subsystem. If it is on the Liouville space, it must
    # exist on a space as large as the square of the Hilbert dimension.
    required_shape = ([x**2 for x in aff_subs_dim_ar[0]] if channel.issuper
                      else aff_subs_dim_ar[0])
    if tuple(required_shape) != channel.shape:
        raise ValueError(
            "superoperator dimension must be the subsystem dimension squared"
        )
    mask = np.asarray(mask)
    if reference:
        return _subsystem_apply_reference(state, channel, mask)
    if state.isoper:
        return _subsystem_apply_dm(state, channel, mask)
    return _subsystem_apply_ket(state, channel, mask)


def _subsystem_apply_ket(state, channel, mask):
    # TODO Write more efficient code for single-matrix map on pure states
    # TODO Write more efficient code for single-subsystem map . . .
    return _subsystem_apply_dm(state.proj(), channel, mask)


def _subsystem_apply_dm(state, channel, mask):
    """
    Applies a channel to every subsystem indicated by a mask, by
    repeatedly applying the channel to each affected subsystem.
    """
    # Initialize Output Matrix
    rho_out = state
    # checked affected subsystems print arange(len(state.dims[0]))[mask]
    for subsystem in np.arange(len(state.dims[0]))[mask]:
        rho_out = _one_subsystem_apply(rho_out, channel, subsystem)
    return rho_out


def _one_subsystem_apply(state, channel, idx):
    """
    Applies a channel to a state on one subsystem, by breaking it into
    blocks and applying a reduced channel to each block.
    """
    subs_dim_ar = np.array(state.dims).T

    # Calculate number of blocks
    n_blks = 1
    for mul_idx in range(idx):
        n_blks = n_blks * subs_dim_ar[mul_idx][0]
    blk_sz = state.shape[0] // n_blks

    # Apply channel to top subsystem of each block in matrix
    full_data_matrix = state.full()
    for blk_r in range(n_blks):
        for blk_c in range(n_blks):
            # Apply channel to high-level blocks of matrix:
            blk_rx = blk_r * blk_sz
            blk_cx = blk_c * blk_sz

            full_data_matrix[blk_rx : blk_rx+blk_sz, blk_cx : blk_cx+blk_sz] =\
                _block_apply(full_data_matrix[blk_rx : blk_rx+blk_sz,
                                              blk_cx : blk_cx+blk_sz],
                             channel)
    return Qobj(full_data_matrix, dims=state.dims)


def _block_apply(block, channel):
    return (_top_apply_U(block, channel) if channel.isoper
            else _top_apply_S(block, channel))


def _top_apply_U(block, channel):
    """
    Uses scalar-matrix multiplication to efficiently apply a channel to
    the leftmost register in the tensor product, given a unitary matrix
    for a channel.
    """
    split_mat = _block_split(block, *channel.shape)
    temp_split_mat = np.zeros(np.shape(split_mat)).astype(complex)
    for dm_row_idx in range(channel.shape[0]):
        for dm_col_idx in range(channel.shape[1]):
            for op_row_idx in range(channel.shape[0]):
                for op_col_idx in range(channel.shape[1]):
                    temp_split_mat[dm_row_idx][dm_col_idx] += (
                        channel[dm_row_idx, op_col_idx]
                        * channel[dm_col_idx, op_row_idx].conjugate()
                        * split_mat[op_col_idx][op_row_idx]
                    )
    return _block_join(temp_split_mat)


def _top_apply_S(block, channel):
    # If the channel is a super-operator, perform second block decomposition;
    # block-size matches Hilbert space of affected subsystem:
    # FIXME use state shape?
    n_v = int(np.sqrt(channel.shape[0]))
    n_h = int(np.sqrt(channel.shape[1]))
    column = _block_col(block, n_v, n_h)
    chan_mat = channel.full()
    temp_col = np.zeros(np.shape(column)).astype(complex)
    for i, row in enumerate(chan_mat):
        temp_col[i] = sum(s * mat for s, mat in zip(row.T, column))
    return _block_stack(temp_col, n_v, n_h)


def _block_split(mat_in, n_v, n_h):
    """
    Returns a 4D array of matrices, splitting mat_in into
    n_v * n_h square sub-arrays.
    """
    return [np.hsplit(x, n_h) for x in np.vsplit(mat_in, n_v)]


def _block_join(mat_in):
    return np.hstack(np.hstack(mat_in))


def _block_col(mat_in, n_v, n_h):
    """
    Returns a 3D array of matrices, splitting mat_in into
    n_v * n_h square sub-arrays.
    """
    rows, cols = np.shape(mat_in)
    return np.reshape(
        np.array(_block_split(mat_in, n_v, n_h)).transpose(1, 0, 2, 3),
        (n_v * n_h, rows // n_v, cols // n_h)
    )


def _block_stack(arr_in, n_v, n_h):
    """
    Inverse of _block_split
    """
    rs, cs = np.shape(arr_in)[-2:]
    temp = list(map(np.transpose, arr_in))
    # print shape(arr_in)
    temp = np.reshape(temp, (n_v, n_h, rs, cs))
    return np.hstack(np.hstack(temp)).T


def _subsystem_apply_reference(state, channel, mask):
    state = state.proj() if state.isket else state
    if channel.isoper:
        full_oper = tensor([channel if mask_ else qeye(size)
                            for size, mask_ in zip(state.dims[0], mask)])
        return full_oper @ state @ full_oper.dag()
    kraus_list = to_kraus(channel)
    # Kraus operators to be padded with identities:
    k_qubit_kraus_list = itertools.product(kraus_list, repeat=np.sum(mask))
    rho_out = Qobj(_data.csr.zeros(*state.shape), dims=state.dims)
    for operator_iter in k_qubit_kraus_list:
        operator_iter = iter(operator_iter)
        op_iter_list = [next(operator_iter) if mask[j]
                        else qeye(state.dims[0][j])
                        for j in range(len(state.dims[0]))]
        full_oper = tensor(list(map(Qobj, op_iter_list)))
        rho_out = rho_out + full_oper * state * full_oper.dag()
    return rho_out



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True

from libc.string cimport memcpy, memset

cimport cython

import warnings

from qutip.core.data.base cimport idxint
from qutip.core.data cimport csr, dense, CSR, Dense, Data, Dia

__all__ = [
    'reshape', 'reshape_csr', 'reshape_dense', 'reshape_dia',
    'column_stack', 'column_stack_csr', 'column_stack_dense', 'column_stack_dia',
    'column_unstack', 'column_unstack_csr', 'column_unstack_dense', 'column_unstack_dia',
    'split_columns', 'split_columns_dense', 'split_columns_csr', 'split_columns_dia',
]


cdef void _reshape_check_input(Data matrix, idxint n_rows_out, idxint n_cols_out) except *:
    if n_rows_out * n_cols_out != matrix.shape[0] * matrix.shape[1]:
        message = "".join([
            "cannot reshape ", str(matrix.shape), " to ",
            "(", str(n_rows_out), ", ", str(n_cols_out), ")",
        ])
        raise ValueError(message)
    if n_rows_out <= 0 or n_cols_out <= 0:
        raise ValueError("must have > 0 rows and columns")


cpdef CSR reshape_csr(CSR matrix, idxint n_rows_out, idxint n_cols_out):
    cdef size_t ptr, row_in, row_out=0, loc, cur=0
    cdef size_t n_rows_in=matrix.shape[0], n_cols_in=matrix.shape[1]
    cdef idxint nnz = csr.nnz(matrix)
    cdef CSR out
    _reshape_check_input(matrix, n_rows_out, n_cols_out)
    out = csr.empty(n_rows_out, n_cols_out, nnz)
    matrix.sort_indices()
    with nogil:
        # Since the indices are now sorted, the data arrays will be identical.
        memcpy(out.data, matrix.data, nnz*sizeof(double complex))
        memset(out.row_index, 0, (n_rows_out + 1) * sizeof(idxint))
        for row_in in range(n_rows_in):
            for ptr in range(matrix.row_index[row_in], matrix.row_index[row_in+1]):
                loc = cur + matrix.col_index[ptr]
                out.row_index[loc // n_cols_out + 1] += 1
                out.col_index[ptr] = loc % n_cols_out
            cur += n_cols_in
        for row_out in range(n_rows_out):
            out.row_index[row_out + 1] += out.row_index[row_out]
    return out


cdef inline idxint _reshape_dense_reindex(idxint idx, idxint size):
    return (idx // size) + (idx % size)


cpdef Dense reshape_dense(Dense matrix, idxint n_rows_out, idxint n_cols_out):
    _reshape_check_input(matrix, n_rows_out, n_cols_out)
    cdef Dense out
    if not matrix.fortran:
        out = matrix.copy()
        out.shape = (n_rows_out, n_cols_out)
        return out
    out = dense.zeros(n_rows_out, n_cols_out)
    cdef size_t idx_in=0, idx_out=0
    cdef size_t size = n_rows_out * n_cols_out
    # TODO: improve the algorithm here.
    cdef size_t stride = _reshape_dense_reindex(matrix.shape[1]*n_rows_out, size)
    for idx_in in range(size):
        out.data[idx_out] = matrix.data[idx_in]
        idx_out = _reshape_dense_reindex(idx_out + stride, size)
    return out


cpdef Dia reshape_dia(Dia matrix, idxint n_rows_out, idxint n_cols_out):
    _reshape_check_input(matrix, n_rows_out, n_cols_out)
    # Once reshaped, diagonals are no longer ligned up.
    return Dia(
        matrix.as_scipy().reshape((n_rows_out, n_cols_out)).todia(), copy=False
    )


cpdef CSR column_stack_csr(CSR matrix):
    if matrix.shape[1] == 1:
        return matrix.copy()
    return reshape_csr(matrix.transpose(), matrix.shape[0]*matrix.shape[1], 1)


cpdef Dense column_stack_dense(Dense matrix, bint inplace=False):
    cdef Dense out
    if inplace and matrix.fortran:
        matrix.shape = (matrix.shape[0] * matrix.shape[1], 1)
        return matrix
    if matrix.fortran:
        out = matrix.copy()
        out.shape = (matrix.shape[0]*matrix.shape[1], 1)
        return out
    if inplace:
        warnings.warn("cannot stack columns inplace for C-ordered matrix")
    return reshape_dense(matrix.transpose(), matrix.shape[0]*matrix.shape[1], 1)


cpdef Dia column_stack_dia(Dia matrix):
    if matrix.shape[1] == 1:
        return matrix.copy()
    return reshape_dia(matrix.transpose(), matrix.shape[0]*matrix.shape[1], 1)


cdef void _column_unstack_check_shape(Data matrix, idxint rows) except *:
    if matrix.shape[1] != 1:
        raise ValueError("input is not a single column")
    if rows < 1:
        raise ValueError("rows must be a positive integer")
    if matrix.shape[0] % rows:
        raise ValueError("number of rows does not divide into the shape")


cpdef CSR column_unstack_csr(CSR matrix, idxint rows):
    _column_unstack_check_shape(matrix, rows)
    cdef idxint cols = matrix.shape[0] // rows
    return reshape_csr(matrix, cols, rows).transpose()


cpdef Dense column_unstack_dense(Dense matrix, idxint rows, bint inplace=False):
    _column_unstack_check_shape(matrix, rows)
    cdef idxint cols = matrix.shape[0] // rows
    if inplace and matrix.fortran:
        matrix.shape = (rows, cols)
        return matrix
    elif inplace:
        warnings.warn("cannot unstack columns inplace for C-ordered matrix")
    out = dense.empty(rows, cols, fortran=True)
    memcpy(out.data, matrix.data, rows*cols * sizeof(double complex))
    return out


cpdef Dia column_unstack_dia(Dia matrix, idxint rows):
    _column_unstack_check_shape(matrix, rows)
    cdef idxint cols = matrix.shape[0] // rows
    return reshape_dia(matrix, cols, rows).transpose()


cpdef list split_columns_dense(Dense matrix, copy=True):
    return [Dense(matrix.as_ndarray()[:, k], copy=copy)
            for k in range(matrix.shape[1])]


cpdef list split_columns_csr(CSR matrix, copy=True):
    return [CSR(matrix.as_scipy()[:, k], copy=copy)
            for k in range(matrix.shape[1])]


cpdef list split_columns_dia(Dia matrix, copy=None):
    as_array = matrix.to_array()
    return [Dense(as_array[:, k], copy=False) for k in range(matrix.shape[1])]


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

reshape = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('n_rows_out', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
        _inspect.Parameter('n_cols_out', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
    ]),
    name='reshape',
    module=__name__,
    inputs=('matrix',),
    out=True,
)
reshape.__doc__ =\
    """
    Reshape the input matrix.  The values of `n_rows_out` and `n_cols_out` must
    match the current total number of elements of the matrix.

    Parameters
    ----------
    matrix : Data
        The input matrix to reshape.

    n_rows_out, n_cols_out : integer
        The number of rows and columns in the output matrix.
    """
reshape.add_specialisations([
    (CSR, CSR, reshape_csr),
    (Dense, Dense, reshape_dense),
    (Dia, Dia, reshape_dia),
], _defer=True)


# Similar to the `out` parameter of `matmul`, we don't include `inplace` in the
# signature of `column_stack` because the dispatcher logic currently doesn't
# support the idea of an input parameter also being the output.

column_stack = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='column_stack',
    module=__name__,
    inputs=('matrix',),
    out=True,
)
column_stack.__doc__ =\
    """
    Stack the columns of a matrix so it becomes a ket-like vector.  For
    example, the matrix
        [[0, 1, 2],
         [3, 4, 5],
         [6, 7, 8]]
    would be transformed to
        [[0],
         [3],
         [6],
         [1],
         ...
         [5],
         [8]]
    This is used for transforming between operator and operator-ket
    representations in the super-operator formalism.

    The inverse of this operation is `column_unstack`.

    Parameters
    ----------
    matrix : Data
        The matrix to stack the columns of.
    """
column_stack.add_specialisations([
    (CSR, CSR, column_stack_csr),
    (Dense, Dense, column_stack_dense),
    (Dia, Dia, column_stack_dia),
], _defer=True)

column_unstack = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('rows', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
    ]),
    name='column_unstack',
    module=__name__,
    inputs=('matrix',),
    out=True,
)
column_unstack.__doc__ =\
    """
    Unstack the columns of a ket-like vector so it becomes a matrix with `rows`
    number of rows.  For example, the matrix
        [[0],
         [1],
         [2],
         [3]]
    would be unstacked with `rows = 2` to
        [[0, 2],
         [1, 3]]
    This is used for transforming between the operator-ket and operator
    representations in the super-operator formalism.

    The inverse of this operation is `column_stack`.

    Parameters
    ----------
    matrix : Data
        The matrix to unstack the columns of.

    rows : integer
        The number of rows there should be in the output matrix.  This must
        divide into the total number of elements in the input.
    """
column_unstack.add_specialisations([
    (CSR, CSR, column_unstack_csr),
    (Dense, Dense, column_unstack_dense),
    (Dia, Dia, column_unstack_dia),
], _defer=True)


split_columns = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('copy', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=1),
    ]),
    name='split_columns',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
split_columns.__doc__ =\
    """
    Make a ket-shaped data out of each columns of the matrix.
    This is used for to split the eigenvectors from :obj:`eigs`.

    Parameters
    ----------
    matrix : Data
        The matrix to unstack the columns of.

    copy : bool, optional [True]
        Whether to return copy of the data or a view.
        View may not be a possible in all cases.

    Returns
    -------
    kets : list
        list of Data of each columns of the matrix.
    """
split_columns.add_specialisations([
    (CSR, split_columns_csr),
    (Dense, split_columns_dense),
    (Dia, split_columns_dia),
], _defer=True)

del _inspect, _Dispatcher



__all__ = ['BaseProgressBar', 'TextProgressBar',
           'EnhancedTextProgressBar', 'TqdmProgressBar',
           'HTMLProgressBar', 'progress_bars']

import time
import datetime
import sys
from qutip import settings


class BaseProgressBar(object):
    """
    An abstract progress bar with some shared functionality.

    Example usage:

        n_vec = linspace(0, 10, 100)
        pbar = TextProgressBar(len(n_vec))
        for n in n_vec:
            pbar.update()
            compute_with_n(n)
        pbar.finished()

    """

    def __init__(self, iterations=0, chunk_size=10, **kwargs):
        self._start(iterations, chunk_size)

    def _start(self, iterations, chunk_size=10, **kwargs):
        self.N = float(iterations)
        self.n = 0
        self.p_chunk_size = chunk_size
        self.p_chunk = chunk_size
        self.t_start = time.time()
        self.t_done = self.t_start - 1

    def update(self):
        pass

    def total_time(self):
        return self.t_done - self.t_start

    def time_elapsed(self):
        return "%6.2fs" % (time.time() - self.t_start)

    def time_remaining_est(self, p):
        if p > 0.0:
            t_r_est = (time.time() - self.t_start) * (100.0 - p) / p
        else:
            t_r_est = 0

        dd = datetime.datetime(1, 1, 1) + datetime.timedelta(seconds=t_r_est)
        time_string = "%02d:%02d:%02d:%02d" % \
            (dd.day - 1, dd.hour, dd.minute, dd.second)

        return time_string

    def finished(self):
        self.t_done = time.time()


class TextProgressBar(BaseProgressBar):
    """
    A simple text-based progress bar.
    """

    def __init__(self, iterations=0, chunk_size=10, **kwargs):
        super()._start(iterations, chunk_size)

    def update(self):
        self.n += 1
        n = self.n
        p = (n / self.N) * 100.0
        if p >= self.p_chunk:
            print("%4.1f%%." % p +
                  " Run time: %s." % self.time_elapsed() +
                  " Est. time left: %s" % self.time_remaining_est(p))
            sys.stdout.flush()
            self.p_chunk += self.p_chunk_size

    def finished(self):
        self.t_done = time.time()
        print("Total run time: %s" % self.time_elapsed())


class EnhancedTextProgressBar(BaseProgressBar):
    """
    An enhanced text-based progress bar.
    """

    def __init__(self, iterations=0, chunk_size=10, **kwargs):
        super()._start(iterations, chunk_size)
        self.fill_char = '*'
        self.width = 25

    def update(self):
        self.n += 1
        n = self.n
        percent_done = int(round(n / self.N * 100.0))
        all_full = self.width - 2
        num_hashes = int(round((percent_done / 100.0) * all_full))
        prog_bar = ('[' + self.fill_char * num_hashes +
                    ' ' * (all_full - num_hashes) + ']')
        pct_place = (len(prog_bar) // 2) - len(str(percent_done))
        pct_string = '%d%%' % percent_done
        prog_bar = (prog_bar[0:pct_place] +
                    (pct_string + prog_bar[pct_place + len(pct_string):]))
        prog_bar += ' Elapsed %s / Remaining %s' % (
            self.time_elapsed().strip(),
            self.time_remaining_est(percent_done))
        print('\r', prog_bar, end='')
        sys.stdout.flush()

    def finished(self):
        self.t_done = time.time()
        print("\r", "Total run time: %s" % self.time_elapsed())


class TqdmProgressBar(BaseProgressBar):
    """
    A progress bar using tqdm module
    """

    def __init__(self, iterations=0, chunk_size=10, **kwargs):
        from tqdm.auto import tqdm
        self.pbar = tqdm(total=iterations, **kwargs)
        self.t_start = time.time()
        self.t_done = self.t_start - 1

    def update(self):
        self.pbar.update()

    def finished(self):
        self.pbar.close()
        self.t_done = time.time()


class HTMLProgressBar(BaseProgressBar):
    """
    A simple HTML progress bar for using in IPython notebooks. Based on
    IPython ProgressBar demo notebook:
    https://github.com/ipython/ipython/tree/master/examples/notebooks

    Example usage:

        n_vec = linspace(0, 10, 100)
        pbar = HTMLProgressBar(len(n_vec))
        for n in n_vec:
            pbar.update()
            compute_with_n(n)
    """

    def __init__(self, iterations=0, chunk_size=1.0, **kwargs):
        from IPython.display import HTML, Javascript, display
        import uuid

        self.display = display
        self.Javascript = Javascript
        self.divid = str(uuid.uuid4())
        self.textid = str(uuid.uuid4())
        self.pb = HTML(
            '<div style="border: 2px solid grey; width: 600px">\n  '
            f'<div id="{self.divid}" '
            'style="background-color: rgba(121,195,106,0.75); '
            'width:0%">&nbsp;</div>\n'
            '</div>\n'
            f'<p id="{self.textid}"></p>\n'
        )
        self.display(self.pb)
        super()._start(iterations, chunk_size)

    def update(self):
        self.n += 1
        n = self.n
        p = (n / self.N) * 100.0
        if p >= self.p_chunk:
            lbl = ("Elapsed time: %s. " % self.time_elapsed() +
                   "Est. remaining time: %s." % self.time_remaining_est(p))
            js_code = ("$('div#%s').width('%i%%');" % (self.divid, p) +
                       "$('p#%s').text('%s');" % (self.textid, lbl))
            self.display(self.Javascript(js_code))
            self.p_chunk += self.p_chunk_size

    def finished(self):
        self.t_done = time.time()
        lbl = "Elapsed time: %s" % self.time_elapsed()
        js_code = ("$('div#%s').width('%i%%');" % (self.divid, 100.0) +
                   "$('p#%s').text('%s');" % (self.textid, lbl))
        self.display(self.Javascript(js_code))


progress_bars = {
    "Enhanced": EnhancedTextProgressBar,
    "enhanced": EnhancedTextProgressBar,
    "Text": TextProgressBar,
    "text": TextProgressBar,
    True: TextProgressBar,
    "Tqdm": TqdmProgressBar,
    "tqdm": TqdmProgressBar,
    "Html": HTMLProgressBar,
    "html": HTMLProgressBar,
    "base": BaseProgressBar,
    "": BaseProgressBar,
    False: BaseProgressBar,
    None: BaseProgressBar,
}



import numpy as np

__all__ = ["Wiener"]


class Wiener:
    """
    Wiener process.
    """
    def __init__(self, t0, dt, generator, shape):
        self.t0 = t0
        self.dt = dt
        self.generator = generator
        self.t_end = t0
        self.shape = shape
        self.process = np.zeros((1,) + shape, dtype=float)

    def _extend(self, t):
        N_new_vals = int((t - self.t_end + self.dt*0.01) // self.dt)
        dW = self.generator.normal(
            0, np.sqrt(self.dt), size=(N_new_vals,) + self.shape
        )
        W = self.process[-1, :, :] + np.cumsum(dW, axis=0)
        self.process = np.concatenate((self.process, W), axis=0)
        self.t_end = self.t0 + (self.process.shape[0] - 1) * self.dt

    def dW(self, t, N):
        if t + N * self.dt > self.t_end:
            self._extend(t + N * self.dt)
        idx0 = int((t - self.t0 + self.dt * 0.01) // self.dt)
        return np.diff(self.process[idx0:idx0 + N + 1, :, :], axis=0)

    def __call__(self, t):
        if t > self.t_end:
            self._extend(t)
        idx = int((t - self.t0 + self.dt * 0.01) // self.dt)
        return self.process[idx, 0, :]


class _Noise:
    """
    Wiener process generator used for tests.
    """

    def __init__(self, T, dt, num=1):
        N = int(np.round(T / dt))
        self.T = T
        self.dt = dt
        self.num = num
        self.noise = np.random.randn(N, num) * dt**0.5

    def dw(self, dt):
        """
        Ito integral I(i).
        """
        N = int(np.round(dt / self.dt))
        return self.noise.reshape(-1, N, self.num).sum(axis=1)

    def dz(self, dt):
        """
        Ito integral I(0, i).
        """
        N = int(np.round(dt / self.dt))
        return (
            np.einsum(
                "ijk,j->ik",
                self.noise.reshape(-1, N, self.num),
                np.arange(N - 0.5, 0, -1),
            )
            * self.dt
        )

    def dW(self, dt):
        """
        Noise used for Ito-Taylor integrators of order up to 1.5.
        """
        N = int(np.round(dt / self.dt))
        noise = self.noise.copy()
        if noise.shape[0] % N:
            noise = noise[: -(noise.shape[0] % N)]
        out = np.empty((noise.shape[0] // N, 2, self.num), dtype=float)
        out[:, 0, :] = noise.reshape(-1, N, self.num).sum(axis=1)
        out[:, 1, :] = (
            np.einsum(
                "ijk,j->ik",
                self.noise.reshape(-1, N, self.num),
                np.arange(N - 0.5, 0, -1),
            )
            * self.dt
        )
        return out



# -*- coding: utf-8 -*-

"""
This module implements internal-use functions for semidefinite programming.
"""
import collections
import functools

import numpy as np
import scipy.sparse as sp

# Conditionally import CVXPY
try:
    import cvxpy

    __all__ = ["dnorm_problem", "dnorm_sparse_problem"]
except ImportError:
    cvxpy = None
    __all__ = []

from .operators import swap

Complex = collections.namedtuple("Complex", ["re", "im"])


def _complex_var(rows=1, cols=1, name=None):
    return Complex(
        re=cvxpy.Variable((rows, cols), name=(name + "_re") if name else None),
        im=cvxpy.Variable((rows, cols), name=(name + "_im") if name else None),
    )


def _make_constraints(*rhos):
    """
    Create constraints to ensure definied density operators.
    """
    # rhos traces are 1
    constraints = [cvxpy.trace(rho.re) == 1 for rho in rhos]
    # rhos are Hermitian
    for rho in rhos:
        constraints += [rho.re == rho.re.T] + [rho.im == -rho.im.T]
    # Non negative
    constraints += [
        cvxpy.bmat([[rho.re, -rho.im], [rho.im, rho.re]]) >> 0 for rho in rhos
    ]
    return constraints


def _arr_to_complex(A):
    if np.iscomplex(A).any():
        return Complex(re=A.real, im=A.imag)
    return Complex(re=A, im=np.zeros_like(A))


def _kron(A, B):
    if isinstance(A, np.ndarray):
        A = _arr_to_complex(A)
    if isinstance(B, np.ndarray):
        B = _arr_to_complex(B)

    return Complex(
        re=(cvxpy.kron(A.re, B.re) - cvxpy.kron(A.im, B.im)),
        im=(cvxpy.kron(A.im, B.re) + cvxpy.kron(A.re, B.im)),
    )


def _conj(W, A):
    U, V = W.re, W.im
    A, B = A.re, A.im
    return Complex(
        re=(U @ A @ U.T - U @ B @ V.T - V @ A @ V.T - V @ B @ U.T),
        im=(U @ A @ V.T + U @ B @ U.T + V @ A @ U.T - V @ B @ V.T),
    )


@functools.lru_cache
def initialize_constraints_on_dnorm_problem(dim):
    # Start assembling constraints and variables.
    constraints = []

    # Make a complex variable for X.
    X = _complex_var(dim**2, dim**2, "X")

    # Make complex variables for rho0 and rho1.
    rho0 = _complex_var(dim, dim, "rho0")
    rho1 = _complex_var(dim, dim, "rho1")
    constraints += _make_constraints(rho0, rho1)

    # Finally, add the tricky positive semidefinite constraint.
    # Since we're using column-stacking, but Watrous used row-stacking,
    # we need to swap the order in Rho0 and Rho1. This is not straightforward,
    # as CVXPY requires that the constant be the first argument. To solve this,
    # We conjugate by SWAP.
    W = swap(dim, dim).full()
    W = Complex(re=W.real, im=W.imag)
    Rho0 = _conj(W, _kron(np.eye(dim), rho0))
    Rho1 = _conj(W, _kron(np.eye(dim), rho1))

    Y = cvxpy.bmat(
        [
            [Rho0.re, X.re, -Rho0.im, -X.im],
            [X.re.T, Rho1.re, X.im.T, -Rho1.im],
            [Rho0.im, X.im, Rho0.re, X.re],
            [-X.im.T, Rho1.im, X.re.T, Rho1.re],
        ]
    )
    constraints += [Y >> 0]

    return X, constraints


def dnorm_problem(dim):
    """
    Creade the cvxpy ``Problem`` for the dnorm metric using dense arrays
    """
    X, constraints = initialize_constraints_on_dnorm_problem(dim)
    Jr = cvxpy.Parameter((dim**2, dim**2))
    Ji = cvxpy.Parameter((dim**2, dim**2))
    # The objective, however, depends on J.
    objective = cvxpy.Maximize(cvxpy.trace(Jr.T @ X.re + Ji.T @ X.im))
    problem = cvxpy.Problem(objective, constraints)
    return problem, Jr, Ji


def dnorm_sparse_problem(dim, J_dat):
    """
    Creade the cvxpy ``Problem`` for the dnorm metric using sparse arrays
    """
    X, constraints = initialize_constraints_on_dnorm_problem(dim)
    J_val = J_dat.tocoo()

    def adapt_sparse_params(A_val, dim):
        # This detour is needed as pointed out in cvxgrp/cvxpy#1159, as cvxpy
        # can not solve with parameters that aresparse matrices directly.
        # Solutions have to be made through calling cvxpy.reshape on
        # the original sparse matrix.
        side_size = dim**2
        A_nnz = cvxpy.Parameter(A_val.nnz)

        A_data = np.ones(A_nnz.size)
        A_rows = A_val.row * side_size + A_val.col
        A_cols = np.arange(A_nnz.size)
        # We are pushing the data on the location of the nonzero elements
        # to the nonzero rows of A_indexer
        A_Indexer = sp.coo_matrix(
            (A_data, (A_rows, A_cols)), shape=(side_size**2, A_nnz.size)
        )
        # We get finaly the sparse matrix A which we wanted
        A = cvxpy.reshape(A_Indexer @ A_nnz, (side_size, side_size), order="C")
        A_nnz.value = A_val.data
        return A

    Jr_val = J_val.real
    Jr = adapt_sparse_params(Jr_val, dim)

    Ji_val = J_val.imag
    Ji = adapt_sparse_params(Ji_val, dim)

    # The objective, however, depends on J.
    objective = cvxpy.Maximize(cvxpy.trace(Jr.T @ X.re + Ji.T @ X.im))

    problem = cvxpy.Problem(objective, constraints)
    return problem



import numpy as np
import scipy.linalg

import qutip
from qutip.core import data as _data
import pytest


def f(t, args):
    return t * (1 - 0.5j)


def liouvillian_ref(H, c_ops=()):
    L = -1.0j * (qutip.spre(H) - qutip.spost(H)) if H else 0
    for c in c_ops:
        if c.issuper:
            L += c
        else:
            cdc = c.dag() * c
            L += qutip.sprepost(c, c.dag())
            L -= 0.5 * (qutip.spre(cdc) + qutip.spost(cdc))
    return L


class TestMatVec:
    """
    A test class for the QuTiP function for matrix/vector conversion.
    """
    def testOperatorVector(self):
        """
        Superoperator: Operator - vector - operator conversion.
        """
        N = 3
        rho1 = qutip.rand_dm(N)
        rho2 = qutip.vector_to_operator(qutip.operator_to_vector(rho1))
        np.testing.assert_allclose(rho1.full(), rho2.full(), 1e-8)

    def testsuperrep(self):
        N = 3
        rho1 = qutip.rand_dm(N)
        as_vec = qutip.operator_to_vector(rho1)
        assert as_vec.superrep == 'super'

        with pytest.raises(TypeError) as err:
            bad_vec = as_vec.copy()
            bad_vec.superrep = "bad"
            qutip.vector_to_operator(bad_vec)
        assert err.value.args[0] == ("only defined for operator-kets "
                                     "in super format")

        with pytest.raises(TypeError) as err:
            qutip.operator_to_vector(as_vec)
        assert err.value.args[0] == ("Cannot convert object already "
                                     "in super representation")

    def testOperatorVectorTensor(self):
        """
        Superoperator: Operator - vector - operator conversion with a tensor
        product state.
        """
        Na = 3
        Nb = 2
        rhoa = qutip.rand_dm(Na)
        rhob = qutip.rand_dm(Nb)
        rho1 = qutip.tensor(rhoa, rhob)
        rho2 = qutip.vector_to_operator(qutip.operator_to_vector(rho1))
        np.testing.assert_allclose(rho1.full(), rho2.full(), 1e-8)

    def testOperatorVectorNotSquare(self):
        """
        Superoperator: Operator - vector - operator conversion for non-square
        matrix.
        """
        op1 = qutip.Qobj(np.random.rand(6).reshape((3, 2)))
        op2 = qutip.vector_to_operator(qutip.operator_to_vector(op1))
        np.testing.assert_allclose(op1.full(), op2.full(), 1e-8)

    def testOperatorSpreAppl(self):
        """
        Superoperator: apply operator and superoperator from left (spre)
        """
        N = 3
        rho = qutip.rand_dm(N)
        U = qutip.rand_unitary(N)
        rho1 = U * rho
        rho2_vec = qutip.spre(U) * qutip.operator_to_vector(rho)
        rho2 = qutip.vector_to_operator(rho2_vec)
        np.testing.assert_allclose(rho1.full(), rho2.full(), 1e-8)

    def testOperatorSpostAppl(self):
        """
        Superoperator: apply operator and superoperator from right (spost)
        """
        N = 3
        rho = qutip.rand_dm(N)
        U = qutip.rand_unitary(N)
        rho1 = rho * U
        rho2_vec = qutip.spost(U) * qutip.operator_to_vector(rho)
        rho2 = qutip.vector_to_operator(rho2_vec)
        np.testing.assert_allclose(rho1.full(), rho2.full(), 1e-8)

    def testOperatorUnitaryTransform(self):
        """
        Superoperator: Unitary transformation with operators and superoperators
        """
        N = 3
        rho = qutip.rand_dm(N)
        U = qutip.rand_unitary(N)
        rho1 = U * rho * U.dag()
        rho2_vec = qutip.sprepost(U, U.dag()) * qutip.operator_to_vector(rho)
        rho2 = qutip.vector_to_operator(rho2_vec)
        np.testing.assert_allclose(rho1.full(), rho2.full(), 1e-8)

    def testMatrixVecMat(self):
        """
        Superoperator: Conversion matrix to vector to matrix
        """
        M = _data.create(np.random.rand(10, 10))
        V = qutip.stack_columns(M)
        M2 = qutip.unstack_columns(V)
        np.testing.assert_allclose(M.to_array(), M2.to_array(), 1e-8)

    def testVecMatVec(self):
        """
        Superoperator: Conversion vector to matrix to vector
        """
        V = _data.create(np.random.rand(100, 1))
        M = qutip.unstack_columns(V)
        V2 = qutip.stack_columns(M)
        np.testing.assert_allclose(V.to_array(), V2.to_array(), 1e-8)

    def testVecMatIndexConversion(self):
        """
        Superoperator: Conversion between matrix and vector indices
        """
        N = 10
        for i in range(N * N):
            assert i == qutip.stacked_index(N, *qutip.unstacked_index(N, i))

    def testVecMatIndexCompability(self):
        """
        Superoperator: Compatibility between matrix/vector and
        corresponding index conversions.
        """
        N = 10
        M = _data.create(np.random.rand(N, N))
        V = qutip.stack_columns(M)
        for idx in range(N * N):
            i, j = qutip.unstacked_index(N, idx)
            assert V.to_array()[idx, 0] == M.to_array()[i, j]

    def test_reshuffle(self):
        U1 = qutip.rand_unitary(2)
        U2 = qutip.rand_unitary(3)
        U3 = qutip.rand_unitary(4)
        U = qutip.tensor(U1, U2, U3)
        S = qutip.to_super(U)
        S_col = qutip.reshuffle(S)
        assert S_col.dims[0] == [[2], [2], [3], [3], [4], [4]]
        assert qutip.reshuffle(S_col) == S

    def test_sprepost(self):
        U1 = qutip.rand_unitary(3)
        U2 = qutip.rand_unitary(3)
        S1 = qutip.spre(U1) * qutip.spost(U2)
        S2 = qutip.sprepost(U1, U2)
        assert S1 == S2

    @pytest.mark.parametrize("dtype", [_data.CSR, _data.Dense])
    @pytest.mark.parametrize("operation", [qutip.spre,
                                           qutip.spost,
                                           qutip.liouvillian,
                                           qutip.lindblad_dissipator])
    def test_operation_dtype(self, dtype, operation):
        """This test checks that the dtype is properly kept with the
        `operation` function."""
        U1 = qutip.rand_unitary(3).to(dtype)
        result = operation(U1)
        assert type(result.data) == dtype

    @pytest.mark.parametrize("dtype", [_data.CSR, _data.Dense])
    def test_sprepost_dtype(self, dtype):
        """This test checks that the dtype is properly kept with `sprepost`
        function."""
        U1 = qutip.rand_unitary(3).to(dtype)
        result = qutip.sprepost(U1, U1)
        assert type(result.data) == dtype

    def testLiouvillianImplem(self):
        """
        Superoperator: Randomized comparison of standard and reference
        Liouvillian functions.
        """
        N1, N2, N3 = 3, 4, 5
        a1 = qutip.tensor(qutip.rand_dm(N1, density=0.75),
                          qutip.qeye([N2, N3]))
        a2 = qutip.tensor(qutip.qeye(N1),
                          qutip.rand_dm(N2, density=0.75),
                          qutip.qeye(N3))
        a3 = qutip.tensor(qutip.qeye([N1, N2]),
                          qutip.rand_dm(N3, density=0.75))
        H = a1.dag()*a1 + a2.dag()*a2 + a3.dag()*a3
        c_ops = [np.sqrt(0.01) * a1, np.sqrt(0.025) * a2, np.sqrt(0.05) * a3]
        L1 = qutip.liouvillian(H, c_ops)
        L2 = liouvillian_ref(H, c_ops)
        np.testing.assert_allclose(L1.full(), L2.full(), 1e-8)


class TestSuper_td:
    """
    A test class for the QuTiP superoperator functions.
    """
    N = 3
    t1 = qutip.QobjEvo([qutip.qeye(N)*(1 + 0.1j),
                        [qutip.create(N) * (1 - 0.1j), f]])
    t2 = qutip.QobjEvo([qutip.destroy(N) * (1 - 0.2j)])
    t3 = qutip.QobjEvo([[qutip.num(N) * (1 + 0.2j), f]])
    q1 = qutip.qeye(N) * (1 + 0.3j)
    q2 = qutip.destroy(N) * (1 - 0.3j)
    q3 = qutip.num(N) * (1 + 0.4j)

    def test_spre_td(self):
        "Superoperator: spre, time-dependent"
        assert qutip.spre(self.t1)(0.5) == qutip.spre(self.t1(0.5))

    def test_spost_td(self):
        "Superoperator: spre, time-dependent"
        assert qutip.spost(self.t1)(0.5) == qutip.spost(self.t1(0.5))

    def test_sprepost_td(self):
        "Superoperator: sprepost, time-dependent"
        # left QobjEvo
        assert (qutip.sprepost(self.t1, self.q2)(0.5)
                == qutip.sprepost(self.t1(0.5), self.q2))
        # left QobjEvo
        assert (qutip.sprepost(self.q2, self.t1)(0.5)
                == qutip.sprepost(self.q2, self.t1(0.5)))
        # left 2 QobjEvo, one cte
        assert (qutip.sprepost(self.t1, self.t2)(0.5)
                == qutip.sprepost(self.t1(0.5), self.t2(0.5)))

    def test_operator_vector_td(self):
        "Superoperator: operator_to_vector, time-dependent"
        assert (qutip.operator_to_vector(self.t1)(0.5)
                == qutip.operator_to_vector(self.t1(0.5)))
        vec = qutip.operator_to_vector(self.t1)
        assert (qutip.vector_to_operator(vec)(0.5)
                == qutip.vector_to_operator(vec(0.5)))

    def test_liouvillian_td(self):
        "Superoperator: liouvillian, time-dependent"
        assert (qutip.liouvillian(self.t1)(0.5)
                == qutip.liouvillian(self.t1(0.5)))
        assert (qutip.liouvillian(None, [self.t2])(0.5)
                == qutip.liouvillian(None, [self.t2(0.5)]))
        assert (qutip.liouvillian(self.t1, [self.t2, self.q1, self.t3],
                                  chi=[1, 2, 3])(0.5)
                == qutip.liouvillian(self.t1(0.5),
                                     [self.t2(0.5), self.q1, self.t3(0.5)],
                                     chi=[1, 2, 3]))

    def test_lindblad_dissipator_td(self):
        "Superoperator: lindblad_dissipator, time-dependent"
        assert (qutip.lindblad_dissipator(self.t2)(0.5)
                == qutip.lindblad_dissipator(self.t2(0.5)))
        assert (qutip.lindblad_dissipator(self.t2, self.q1)(0.5)
                == qutip.lindblad_dissipator(self.t2(0.5), self.q1))
        assert (qutip.lindblad_dissipator(self.q1, self.t2)(0.5)
                == qutip.lindblad_dissipator(self.q1, self.t2(0.5)))



# -*- coding: utf-8 -*-
# @author: Arne L. Grimsmo
# @email1: arne.grimsmo@gmail.com
# @organization: University of Sherbrooke

"""
This module contains an implementation of the non-Markovian transfer tensor
method (TTM), introduced in [1].

[1] Javier Cerrillo and Jianshu Cao, Phys. Rev. Lett 112, 110401 (2014)
"""

import numpy as np
import time
from qutip import spre, vector_to_operator, operator_to_vector, Result


def ttmsolve(dynmaps, state0, times, e_ops=(), num_learning=0, options=None):
    """
    Expand time-evolution using the Transfer Tensor Method [1]_, based on a set
    of precomputed dynamical maps.

    Parameters
    ----------
    dynmaps : list of :class:`.Qobj`, callable
        List of precomputed dynamical maps (superoperators) for the first times
        of ``times`` or a callback function that returns the superoperator at a
        given time.

    state0 : :class:`.Qobj`
        Initial density matrix or state vector (ket).

    times : array_like
        List of times :math:`t_n` at which to compute results.
        Must be uniformily spaced.

    e_ops : :class:`.Qobj`, callable, or list, optional
        Single operator or list of operators for which to evaluate
        expectation values or callable or list of callable.
        Callable signature must be, `f(t: float, state: Qobj)`.
        See :func:`expect` for more detail of operator expectation.

    num_learning : int, default: 0
        Number of times used to construct the dynmaps operators when
        ``dynmaps`` is a callable.

    options : dictionary, optional
        Dictionary of options for the solver.

        - store_final_state : bool
          Whether or not to store the final state of the evolution in the
          result class.
        - store_states : bool, None
          Whether or not to store the state vectors or density matrices.
          On `None` the states will be saved if no expectation operators are
          given.
        - normalize_output : bool
          Normalize output state to hide ODE numerical errors.
        - threshold : float
          Threshold for halting. Halts if  :math:`||T_{n}-T_{n-1}||` is below
          treshold.

    Returns
    -------
    output: :class:`.Result`
        An instance of the class :class:`.Result`.

    .. [1] Javier Cerrillo and Jianshu Cao, Phys. Rev. Lett 112, 110401 (2014)
    """

    opt = {
        "store_final_state": False,
        "store_states": None,
        "normalize_output": "ket",
        "threshold": 0.0,
        "num_learning": 0,
    }
    if options:
        opt.update(options)

    if not np.allclose(np.diff(times), times[1] - times[0]):
        raise ValueError("The time should be uniformily distributed.")

    if callable(dynmaps):
        if num_learning <= 0:
            raise ValueError(
                "When dynmaps is a callable, options['num_learning'] must be "
                "the number of dynamical maps to compute."
            )
        dynmaps = [dynmaps(t) for t in times[:num_learning]]

    if (
        not dynmaps
        or not dynmaps[0].issuper
        or not all(dmap.dims == dynmaps[0].dims for dmap in dynmaps)
    ):
        raise ValueError("`dynmaps` entries must be super operators.")

    start = time.time()
    tensors, diff = _generatetensors(dynmaps, opt["threshold"])
    end = time.time()

    stats = {
        "preparation time": end - start,
        "run time": 0.0,
        "ttmconvergence": diff,
        "num_tensor": len(tensors),
    }
    if state0.isket:
        state0 = state0.proj()

    if state0.isoper:
        # vectorize density matrix
        rho0vec = operator_to_vector(state0)
        restore = vector_to_operator
    else:
        # state0 might be a super in which case we should not vectorize
        rho0vec = state0
        restore = lambda state: state

    K = len(tensors)
    start = time.time()
    results = Result(e_ops, opt, solver="ttmsolve", stats=stats)
    states = [rho0vec]
    results.add(times[0], state0)

    for n in range(1, len(times)):
        # Set current state
        state = 0
        for j in range(1, min(K, n + 1)):
            tmp = tensors[j] @ states[n - j]
            state = tmp + state
        # Append state to all states
        states.append(state)
        results.add(times[n], restore(state))
    end = time.time()
    stats["run time"] = end - start

    return results


def _generatetensors(dynmaps, threshold):
    r"""
    Generate the tensors :math:`T_1,\dots,T_K` from the dynamical maps
    :math:`E(t_k)`.

    A stationary process is assumed, i.e., :math:`T_{n,k} = T_{n-k}`.

    Parameters
    ----------
    dynmaps : list of :class:`.Qobj`
        List of precomputed dynamical maps (superoperators) at the times
        specified in `learningtimes`.

    threshold : float
        Threshold for halting. Halts if  :math:`||T_{n}-T_{n-1}||` is below
        treshold.

    Returns
    -------
    Tensors, diffs: list of :class:`.Qobj.`
        A list of transfer tensors :math:`T_1,\dots,T_K`
    """
    Tensors = []
    diff = [0.0]
    for n in range(len(dynmaps)):
        T = dynmaps[n]
        for m in range(1, n):
            T -= Tensors[n - m] @ dynmaps[m]
        Tensors.append(T)
        if n > 1:
            diff.append((Tensors[-1] - Tensors[-2]).norm())
            if diff[-1] < threshold:
                # Below threshold for truncation
                break
    return Tensors, diff



#cython: language_level=3

from qutip.core.data cimport CSR, Dense, Dia

cpdef double complex trace_csr(CSR matrix) except * nogil
cpdef double complex trace_dense(Dense matrix) except * nogil
cpdef double complex trace_dia(Dia matrix) except * nogil

cpdef double complex trace_oper_ket_csr(CSR matrix) except * nogil
cpdef double complex trace_oper_ket_dense(Dense matrix) except * nogil
cpdef double complex trace_oper_ket_dia(Dia matrix) except * nogil



#cython: language_level=3

from qutip.core.data cimport Dense, Data
from qutip.core.data.base cimport idxint

cdef class QobjEvo:
    cdef:
        list elements
        readonly object _dims
        readonly (idxint, idxint) shape
        int _issuper
        int _isoper
        readonly dict _feedback_functions
        readonly dict _solver_only_feedback

    cpdef Data _call(QobjEvo self, double t)

    cdef object _prepare(QobjEvo self, object t, Data state=*)

    cpdef object expect_data(QobjEvo self, object t, Data state)

    cdef double complex _expect_dense(QobjEvo self, double t, Dense state) except *

    cpdef Data matmul_data(QobjEvo self, object t, Data state, Data out=*)



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport cython
from libc.math cimport sqrt

from qutip.core.data cimport Data, CSR, Dense, Dia
from qutip.core.data cimport base
from .reshape import column_unstack

__all__ = [
    'trace', 'trace_csr', 'trace_dense', 'trace_dia',
    'trace_oper_ket', 'trace_oper_ket_csr', 'trace_oper_ket_dense',
    'trace_oper_ket_dia', 'trace_oper_ket_data',
]


cdef int _check_shape(Data matrix) except -1 nogil:
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("".join([
            "matrix shape ", str(matrix.shape), " is not square.",
        ]))
    return 0


cdef int _check_shape_oper_ket(int N, Data matrix) except -1 nogil:
    if matrix.shape[0] != N * N or matrix.shape[1] != 1:
        raise ValueError("".join([
            "matrix ", str(matrix.shape), " is not a stacked square matrix."
        ]))
    return 0


cpdef double complex trace_csr(CSR matrix) except * nogil:
    _check_shape(matrix)
    cdef size_t row, ptr
    cdef double complex trace = 0
    for row in range(matrix.shape[0]):
        for ptr in range(matrix.row_index[row], matrix.row_index[row + 1]):
            if matrix.col_index[ptr] == row:
                trace += matrix.data[ptr]
                break
    return trace

cpdef double complex trace_dense(Dense matrix) except * nogil:
    _check_shape(matrix)
    cdef double complex trace = 0
    cdef size_t ptr = 0
    cdef size_t stride = matrix.shape[0] + 1
    for _ in range(matrix.shape[0]):
        trace += matrix.data[ptr]
        ptr += stride
    return trace

cpdef double complex trace_dia(Dia matrix) except * nogil:
    _check_shape(matrix)
    cdef double complex trace = 0
    cdef size_t diag, j
    for diag in range(matrix.num_diag):
        if matrix.offsets[diag] == 0:
            for j in range(matrix.shape[1]):
                trace += matrix.data[diag * matrix.shape[1] + j]
            break
    return trace


cpdef double complex trace_oper_ket_csr(CSR matrix) except * nogil:
    cdef size_t N = <size_t>sqrt(matrix.shape[0])
    _check_shape_oper_ket(N, matrix)
    cdef size_t row
    cdef double complex trace = 0
    cdef size_t stride = N + 1
    for row in range(N):
        if matrix.row_index[row * stride] != matrix.row_index[row * stride + 1]:
            trace += matrix.data[matrix.row_index[row * stride]]
    return trace

cpdef double complex trace_oper_ket_dense(Dense matrix) except * nogil:
    cdef size_t N = <size_t>sqrt(matrix.shape[0])
    _check_shape_oper_ket(N, matrix)
    cdef double complex trace = 0
    cdef size_t ptr = 0
    cdef size_t stride = N + 1
    for ptr in range(N):
        trace += matrix.data[ptr * stride]
    return trace


cpdef double complex trace_oper_ket_dia(Dia matrix) except * nogil:
    cdef size_t N = <size_t>sqrt(matrix.shape[0])
    _check_shape_oper_ket(N, matrix)
    cdef double complex trace = 0
    cdef size_t diag = 0
    cdef size_t stride = N + 1
    for diag in range(matrix.num_diag):
        if -matrix.offsets[diag] % stride == 0:
            trace += matrix.data[diag * matrix.shape[1]]
    return trace


cpdef trace_oper_ket_data(Data matrix):
    cdef size_t N = <size_t>sqrt(matrix.shape[0])
    _check_shape_oper_ket(N, matrix)
    return trace(column_unstack(matrix, N))


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

trace = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='trace',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
trace.__doc__ =\
    """Compute the trace (sum of digaonal elements) of a square matrix."""
trace.add_specialisations([
    (CSR, trace_csr),
    (Dia, trace_dia),
    (Dense, trace_dense),
], _defer=True)

trace_oper_ket = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='trace_oper_ket',
    module=__name__,
    inputs=('matrix',),
    out=False,
)
trace_oper_ket.__doc__ =\
    """Compute the trace (sum of digaonal elements) of a stacked square matrix ."""
trace_oper_ket.add_specialisations([
    (CSR, trace_oper_ket_csr),
    (Dia, trace_oper_ket_dia),
    (Dense, trace_oper_ket_dense),
    (Data, trace_oper_ket_data),
], _defer=True)

del _inspect, _Dispatcher



from qutip.core.cy.qobjevo import QobjEvo, _Feedback
from qutip.core.dimensions import Dimensions, Field, SuperSpace
import qutip.core.data as _data
from qutip.core.qobj import Qobj
import numpy as np


class _ExpectFeedback(_Feedback):
    def __init__(self, oper, default=0.):
        self.oper = QobjEvo(oper)
        self.N = oper.shape[1]
        self.N2 = oper.shape[1]**2
        self.default = default

    def check_consistency(self, dims):
        if not (
            self.oper._dims == dims
            or self.oper._dims[1] == dims  # super e_op, oper QobjEvo
            or self.oper._dims == dims[0]  # oper e_op, super QobjEvo
        ):
            raise ValueError(
                f"Dimensions of the expect operator ({self.oper.dims}) "
                f"does not match the operator ({dims})."
            )

    def __call__(self, t, state):
        if state.shape[0] == self.N:
            return self.oper.expect_data(t, state)
        if state.shape[0] == self.N2 and state.shape[1] == 1:
            return self.oper.expect_data(
                t, _data.column_unstack(state, self.N)
            )
        raise ValueError(
            f"Shape of the expect operator ({self.oper.shape}) "
            f"does not match the state ({state.shape})."
        )

    def __repr__(self):
        return "ExpectFeedback"


class _QobjFeedback(_Feedback):
    def __init__(self, default=None, prop=False, open=True):
        self.open = open
        self.prop = prop
        self.default = default

    def check_consistency(self, dims):
        if not self.open:
            # Ket
            self.dims = Dimensions(Field(), dims[1])
        elif dims.issuper:
            # Density matrix, operator already super
            self.dims = dims[1].oper
        else:
            # operator not super, dm dims match oper
            self.dims = dims
        if self.prop and self.dims[1] == Field():
            self.dims = Dimensions(self.dims[0], self.dims[0])
        elif self.prop:
            self.dims = Dimensions(
                SuperSpace(self.dims), SuperSpace(self.dims)
            )

        if self.default is None:
            pass
        elif self.default._dims == self.dims:
            pass
        elif self.default._dims[0] == self.dims[0]:
            # Catch rectangular state and propagator when the flag is not set.
            self.dims = self.default._dims
        else:
            # The state could not be used for qevo @ default...
            raise TypeError(
                f"The dimensions of the default state ({self.default.dims}) "
                f"does not match the operators ({self.dims})."
            )

    def __call__(self, t, state):
        if state.shape == self.dims.shape:
            out = Qobj(state, dims=self.dims)
        else:
            out = Qobj(
                _data.column_unstack(state, self.dims.shape[0]),
                dims=self.dims
            )
        return out

    def __repr__(self):
        return "QobjFeedback"


class _DataFeedback(_Feedback):
    def __init__(self, default=None, open=True, prop=False):
        self.open = open
        self.default = default
        self.prop = prop

    def check_consistency(self, dims):
        if self.default is None:
            return
        if not (
            dims.shape[1] == self.default.shape[0]
            or (dims.shape[1]**2 == self.default.shape[0] and self.open)
        ):
            raise ValueError(
                f"The shape of the default state {self.default.shape} "
                f"does not match the operators {dims.shape}."
            )
        if self.prop and self.default.shape[0] != self.default.shape[1]:
            raise ValueError(
                f"The default state is expected to be square when computing "
                f"propagators, but is {self.default.shape}."
            )

    def __call__(self, t, state):
        return state

    def __repr__(self):
        return "DataFeedback"


class _CollapseFeedback(_Feedback):
    code = "CollapseFeedback"

    def __init__(self, default=None):
        self.default = default or []

    def check_consistency(self, dims):
        pass

    def __repr__(self):
        return "CollapseFeedback"


def _default_weiner(t):
    return np.zeros(1)


class _WeinerFeedback(_Feedback):
    code = "WeinerFeedback"

    def __init__(self, default=None):
        self.default = default or _default_weiner

    def check_consistency(self, dims):
        pass

    def __repr__(self):
        return "WeinerFeedback"



import numpy as np
import pytest

from qutip.core import data
from qutip.core.data import dense, csr

from . import conftest


# Set up some fixtures for automatic parametrisation.

@pytest.fixture(params=[
    pytest.param((1, 5), id='ket'),
    pytest.param((5, 1), id='bra'),
    pytest.param((5, 5), id='square'),
    pytest.param((2, 4), id='wide'),
    pytest.param((4, 2), id='tall'),
])
def shape(request): return request.param


@pytest.fixture(params=[True, False], ids=['Fortran', 'C'])
def fortran(request): return request.param


def _valid_numpy():
    # Arbitrary valid numpy array.
    return conftest.random_numpy_dense((5, 5), False)


@pytest.fixture(scope='function')
def numpy_dense(shape, fortran):
    return conftest.random_numpy_dense(shape, fortran)


@pytest.fixture(scope='function')
def data_dense(shape, fortran):
    return conftest.random_dense(shape, fortran)


class TestClassMethods:
    def test_init_from_ndarray(self, numpy_dense):
        test = data.Dense(numpy_dense)
        assert test.shape == numpy_dense.shape
        assert np.all(test.as_ndarray() == numpy_dense)

    @pytest.mark.parametrize('dtype', ['complex128',
                                       'float64',
                                       'int32', 'int64',
                                       'uint32'])
    def test_init_from_ndarray_other_dtype(self, shape, dtype):
        numpy_dense = np.random.rand(*shape).astype(dtype, casting='unsafe')
        test = data.Dense(numpy_dense)
        assert test.shape == shape
        assert test.as_ndarray().dtype == np.complex128
        assert np.all(test.as_ndarray() == numpy_dense)

    @pytest.mark.parametrize(['arg', 'kwargs', 'error'], [
        pytest.param(_valid_numpy(), {'shape': ()}, ValueError,
                     id="numpy-shape 0 tuple"),
        pytest.param(_valid_numpy(), {'shape': (1,)}, ValueError,
                     id="numpy-shape 1 tuple"),
        pytest.param(_valid_numpy(), {'shape': (None, None)}, ValueError,
                     id="numpy-shape None tuple"),
        pytest.param(_valid_numpy(), {'shape': [2, 2]}, ValueError,
                     id="numpy-shape list"),
        pytest.param(_valid_numpy(), {'shape': (1, 2, 3)}, ValueError,
                     id="numpy-shape 3 tuple"),
        pytest.param(_valid_numpy(), {'shape': (-1, 1)}, ValueError,
                     id="numpy-negative shape"),
        pytest.param(_valid_numpy(), {'shape': (-4, -4)}, ValueError,
                     id="numpy-both negative shape"),
        pytest.param(_valid_numpy(), {'shape': (1213, 1217)}, ValueError,
                     id="numpy-different shape"),
    ])
    def test_init_from_wrong_input(self, arg, kwargs, error):
        """
        Test that the __init__ method raises a suitable error when passed
        incorrectly formatted inputs.

        This test also serves as a *partial* check that Dense safely handles
        deallocation in the presence of exceptions in its __init__ method.  If
        the tests segfault, it's quite likely that the memory management isn't
        being done correctly in the hand-off us setting our data buffers up and
        marking the numpy actually owns the data.
        """
        with pytest.raises(error):
            data.Dense(arg, **kwargs)

    def test_copy_returns_a_correct_copy(self, data_dense):
        """
        Test that the copy() method produces an actual copy, and that the
        result represents the same matrix.
        """
        original = data_dense
        copy = data_dense.copy()
        assert original is not copy
        assert np.all(original.as_ndarray() == copy.as_ndarray())

    def test_as_ndarray_returns_a_view(self, data_dense):
        """
        Test that modifying the views in the result of as_ndarray() also
        modifies the underlying data structures.  This is important for
        allowing data modification from within Python-space.
        """
        unmodified_copy = data_dense.copy()
        data_dense.as_ndarray()[0, 0] += 1
        modified_copy = data_dense.copy()
        assert np.any(data_dense.as_ndarray() != unmodified_copy.as_ndarray())
        assert np.all(data_dense.as_ndarray() == modified_copy.as_ndarray())

    def test_as_ndarray_caches_result(self, data_dense):
        """
        Test that the as_ndarray() method always returns the same view, even if
        called multiple times.
        """
        assert data_dense.as_ndarray() is data_dense.as_ndarray()

    def test_as_ndarray_of_dense_from_ndarray_is_different(self, numpy_dense):
        """
        Test that we produce a new ndarray, regardless of how we have
        initialised the type.
        """
        assert data.Dense(numpy_dense).as_ndarray() is not numpy_dense

    def test_as_ndarray_of_copy_is_different(self, data_dense):
        """
        Test that as_ndarray() does not return the same array or a view to the
        same data, if it's not the same input matrix.  We don't want two Dense
        matrices to be linked.
        """
        original = data_dense.as_ndarray()
        copy = data_dense.copy().as_ndarray()
        assert original is not copy
        assert not np.may_share_memory(original, copy)

    def test_as_ndarray_is_correct_result(self, numpy_dense):
        """
        Test that as_ndarray is actually giving the matrix we expect for a given
        input.
        """
        data_dense = data.Dense(numpy_dense)
        nd_view = data_dense.as_ndarray()
        assert isinstance(data_dense.as_ndarray(), np.ndarray)
        assert nd_view.ndim == 2
        assert nd_view.shape == numpy_dense.shape
        assert nd_view.strides == numpy_dense.strides
        assert np.all(nd_view == numpy_dense)
        assert nd_view.flags.c_contiguous == numpy_dense.flags.c_contiguous
        assert nd_view.flags.f_contiguous == numpy_dense.flags.f_contiguous

    def test_to_array_is_correct_result(self, data_dense):
        test_array = data_dense.to_array()
        nd_view = data_dense.as_ndarray()
        assert isinstance(test_array, np.ndarray)
        assert test_array.ndim == 2
        assert test_array.shape == nd_view.shape
        assert test_array.strides == nd_view.strides
        assert test_array.flags.c_contiguous == nd_view.flags.c_contiguous
        assert test_array.flags.f_contiguous == nd_view.flags.f_contiguous
        # It's not enough to be accurate within a tolerance here - there's no
        # mathematics, so they should be _identical_.
        assert np.all(test_array == nd_view)

    @pytest.mark.parametrize('new_fortran', [
        pytest.param(-1, id='swap'),
        pytest.param(False, id='C'),
        pytest.param(True, id='Fortran'),
    ])
    def test_reorder(self, data_dense, new_fortran):
        reordered = data_dense.reorder(new_fortran)
        assert isinstance(reordered, data.Dense)
        assert reordered.shape == data_dense.shape
        orig = data_dense.to_array()
        test = reordered.to_array()
        if new_fortran == -1:
            orig.flags.f_contiguous == test.flags.c_contiguous
            orig.flags.c_contiguous == test.flags.f_contiguous
        elif new_fortran is True:
            assert test.flags.f_contiguous
        else:
            assert test.flags.c_contiguous
        assert np.all(orig == test)

class TestFactoryMethods:
    def test_empty(self, shape):
        base = dense.empty(shape[0], shape[1])
        nd = base.as_ndarray()
        assert isinstance(base, data.Dense)
        assert base.shape == shape
        assert nd.shape == shape

    def test_zeros(self, shape):
        base = dense.zeros(shape[0], shape[1])
        nd = base.as_ndarray()
        assert isinstance(base, data.Dense)
        assert base.shape == shape
        assert nd.shape == shape
        assert np.count_nonzero(nd) == 0

    @pytest.mark.parametrize('dimension', [1, 5, 100])
    @pytest.mark.parametrize(
        'scale',
        [None, 2, -0.1, 1.5, 1.5+1j],
        ids=['none', 'int', 'negative', 'float', 'complex']
    )
    def test_identity(self, dimension, scale):
        # scale=None is testing that the default value returns the identity.
        base = (dense.identity(dimension) if scale is None
                else dense.identity(dimension, scale))
        nd = base.as_ndarray()
        numpy_test = np.eye(dimension, dtype=np.complex128)
        if scale is not None:
            numpy_test *= scale
        assert isinstance(base, data.Dense)
        assert base.shape == (dimension, dimension)
        assert np.count_nonzero(nd - numpy_test) == 0

    @pytest.mark.parametrize(['diagonals', 'offsets', 'shape'], [
        pytest.param([2j, 3, 5, 9], None, None, id='main diagonal'),
        pytest.param([1], None, None, id='1x1'),
        pytest.param([[0.2j, 0.3]], None, None, id='main diagonal list'),
        pytest.param([0.2j, 0.3], 2, None, id='superdiagonal'),
        pytest.param([0.2j, 0.3], -2, None, id='subdiagonal'),
        pytest.param([[0.2, 0.3, 0.4], [0.1, 0.9]], [-2, 3], None,
                     id='two diagonals'),
        pytest.param([1, 2, 3], 0, (3, 5), id='main wide'),
        pytest.param([1, 2, 3], 0, (5, 3), id='main tall'),
        pytest.param([[1, 2, 3], [4, 5]], [-1, -2], (4, 8), id='two wide sub'),
        pytest.param([[1, 2, 3, 4], [4, 5, 4j, 1j]], [1, 2], (4, 8),
                     id='two wide super'),
        pytest.param([[1, 2, 3], [4, 5]], [1, 2], (8, 4), id='two tall super'),
        pytest.param([[1, 2, 3, 4], [4, 5, 4j, 1j]], [-1, -2], (8, 4),
                     id='two tall sub'),
        pytest.param([[1, 2, 3], [4, 5, 6], [1, 2]], [1, -1, -2], (4, 4),
                     id='out of order'),
        pytest.param([[1, 2, 3], [4, 5, 6], [1, 2]], [1, 1, -2], (4, 4),
                     id='sum duplicates'),
    ])
    def test_diags(self, diagonals, offsets, shape):
        base = dense.diags(diagonals, offsets, shape)
        # Build numpy version test.
        if not isinstance(diagonals[0], list):
            diagonals = [diagonals]
        offsets = np.atleast_1d(offsets if offsets is not None else [0])
        if shape is None:
            size = len(diagonals[0]) + abs(offsets[0])
            shape = (size, size)
        test = np.zeros(shape, dtype=np.complex128)
        for diagonal, offset in zip(diagonals, offsets):
            test[np.where(np.eye(*shape, k=offset) == 1)] += diagonal
        assert isinstance(base, data.Dense)
        assert base.shape == shape
        np.testing.assert_allclose(base.to_array(), test, rtol=1e-10)

    @pytest.mark.parametrize(['shape', 'position', 'value'], [
        pytest.param((1, 1), (0, 0), None, id='minimal'),
        pytest.param((10, 10), (5, 5), 1.j, id='on diagonal'),
        pytest.param((10, 10), (1, 5), 1., id='upper'),
        pytest.param((10, 10), (5, 1), 2., id='lower'),
        pytest.param((10, 1), (5, 0), None, id='column'),
        pytest.param((1, 10), (0, 5), -5.j, id='row'),
        pytest.param((10, 2), (5, 1), 1+2j, id='tall'),
        pytest.param((2, 10), (1, 5), 10, id='wide'),
    ])
    def test_one_element(self, shape, position, value):
        test = np.zeros(shape, dtype=np.complex128)
        if value is None:
            base = data.one_element_dense(shape, position)
            test[position] = 1.0+0.0j
        else:
            base = data.one_element_dense(shape, position, value)
            test[position] = value
        assert isinstance(base, data.Dense)
        assert base.shape == shape
        assert np.allclose(base.to_array(), test, atol=1e-10)

    @pytest.mark.parametrize(['shape', 'position', 'value'], [
        pytest.param((0, 0), (0, 0), None, id='zero shape'),
        pytest.param((10, -2), (5, 0), 1.j, id='neg shape'),
        pytest.param((10, 10), (10, 5), 1., id='outside'),
        pytest.param((10, 10), (5, -1), 2., id='outside neg'),
    ])
    def test_one_element_error(self, shape, position, value):
        with pytest.raises(ValueError) as exc:
            base = data.one_element_dense(shape, position, value)
        assert str(exc.value).startswith("Position of the elements"
                                         " out of bound: ")


def test_OrderEfficiencyWarning():
    N = 5
    M = csr.identity(N)
    C_ordered = dense.zeros(N, 1, fortran=False)
    fortran_ordered = dense.zeros(N, 1, fortran=True)
    with pytest.warns(dense.OrderEfficiencyWarning):
        data.matmul_csr_dense_dense(M, C_ordered, out=fortran_ordered)


@pytest.mark.parametrize("fortran", [True, False])
@pytest.mark.parametrize("func",
    [data.zeros_like_dense, data.identity_like_dense]
)
def test_like_keep_order(func, fortran):
    old = dense.zeros(3, 3, fortran=fortran)
    new = func(old)
    assert new.fortran == old.fortran



import numpy as np
import pytest

from qutip import data as _data
from qutip import CoreOptions

@pytest.fixture(params=[_data.CSR, _data.Dense, _data.Dia], ids=["CSR", "Dense", "Dia"])
def datatype(request):
    return request.param


class Test_isherm:
    tol = 1e-12

    @pytest.mark.repeat(20)
    @pytest.mark.parametrize("density", (0.1, 0.8))
    @pytest.mark.parametrize("size", (10, 100))
    def test_random_equal_structure(self, datatype, size, density):
        # This is only going to be approximate, as entries generated onto the
        # diagonal will fill only one slot, not two, after addition with the
        # transpose.  The densities are arbitrary, though, so this doesn't
        # matter much.
        nnz = int(0.5 * size*size * density) or 1
        indices = np.triu_indices(size)
        choice = np.random.choice(indices[0].size, nnz, replace=False)
        indices = (indices[0][choice], indices[1][choice])

        # Real-symmetric matrices.
        base = np.zeros((size, size), dtype=np.complex128)
        base[indices] = np.random.rand(nnz)
        base += base.T.conj()
        base = _data.to(datatype, _data.create(base))
        assert _data.isherm(base)

        # Complex Hermitian matrices
        base = np.zeros((size, size), dtype=np.complex128)
        base[indices] = np.random.rand(nnz) + 1j*np.random.rand(nnz)
        base += base.T.conj()
        base = _data.to(datatype, _data.create(base))
        assert _data.isherm(base)

        # Complex skew-Hermitian matrices
        base = np.zeros((size, size), dtype=np.complex128)
        base[indices] = np.random.rand(nnz) + 1j*np.random.rand(nnz)
        base += base.T
        base = _data.to(datatype, _data.create(base))
        assert not _data.isherm(base)

    @pytest.mark.parametrize("cols", (2, 4))
    @pytest.mark.parametrize("rows", (1, 5))
    def test_nonsquare_shapes(self, datatype, rows, cols):
        real = _data.to(datatype, _data.create(np.random.rand(rows, cols)))
        assert not _data.isherm(real, self.tol)
        assert not _data.isherm(real.transpose(), self.tol)

        imag = _data.to(
            datatype,
            _data.create(np.random.rand(rows, cols) + 1j * np.random.rand(rows, cols)),
        )
        assert not _data.isherm(imag, self.tol)
        assert not _data.isherm(imag.transpose(), self.tol)

    def test_diagonal_elements(self, datatype):
        n = 10
        base = _data.to(datatype, _data.create(np.diag(np.random.rand(n))))
        assert _data.isherm(base, tol=self.tol)
        assert not _data.isherm(_data.mul(base, 1j), tol=self.tol)

    def test_compare_implicit_zero_structure(self, datatype):
        """
        Regression test for gh-1350, comparing explicitly stored values in the
        matrix (but below the tolerance for allowable Hermicity) to implicit
        zeros.
        """

        with CoreOptions(auto_tidyup=False):
            base = _data.to(
                datatype,
                _data.create(np.array([[1, self.tol * 1e-3j], [0, 1]]))
            )
        # If this first line fails, the zero has been stored explicitly and so
        # the test is invalid.
        assert np.count_nonzero(base.to_array()) == 3
        assert _data.isherm(base, tol=self.tol)
        assert _data.isherm(base.transpose(), tol=self.tol)

        # A similar test if the structures are different, but it's not
        # Hermitian.
        base = _data.to(datatype, _data.create(np.array([[1, 1j], [0, 1]])))
        assert np.count_nonzero(base.to_array()) == 3
        assert not _data.isherm(base, tol=self.tol)
        assert not _data.isherm(base.transpose(), tol=self.tol)

        # Catch possible edge case where it shouldn't be Hermitian, but faulty
        # loop logic doesn't fully compare all rows.
        base = _data.to(
            datatype,
            _data.create(
                np.array(
                    [
                        [0, 0, 0, 0],
                        [0, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 0, 0],
                    ],
                    dtype=np.complex128,
                )
            ),
        )
        assert np.count_nonzero(base.to_array()) == 1
        assert not _data.isherm(base, tol=self.tol)
        assert not _data.isherm(base.transpose(), tol=self.tol)

    @pytest.mark.parametrize("density", np.linspace(0.2, 1, 17))
    def test_compare_implicit_zero_random(self, datatype, density):
        """
        Regression test of gh-1350.

        Larger matrices where all off-diagonal elements are below the
        absolute tolerance, so everything should always appear Hermitian, but
        with random patterns of non-zero elements.  It doesn't matter that it
        isn't Hermitian if scaled up; everything is below absolute tolerance,
        so it should appear so.  We also set the diagonal to be larger to the
        tolerance to ensure isherm can't just compare everything to zero.
        """
        n = 10
        base = self.tol * 1e-2 * (np.random.rand(n, n) + 1j * np.random.rand(n, n))
        # Mask some values out to zero.
        base[np.random.rand(n, n) > density] = 0
        np.fill_diagonal(base, self.tol * 1000)
        nnz = np.count_nonzero(base)
        with CoreOptions(auto_tidyup=False):
            base = _data.to(datatype, _data.create(base))
        assert np.count_nonzero(base.to_array()) == nnz
        assert _data.isherm(base, tol=self.tol)
        assert _data.isherm(base.transpose(), tol=self.tol)

        # Similar test when it must be non-Hermitian.  We set the diagonal
        # to be real because we want to test off-diagonal implicit zeros,
        # and having an imaginary first element would automatically fail.
        nnz = 0
        while nnz <= n:
            # Ensure that we don't just have the real diagonal.
            base = self.tol * 1000j * np.random.rand(n, n)
            # Mask some values out to zero.
            base[np.random.rand(n, n) > density] = 0
            np.fill_diagonal(base, self.tol * 1000)
            nnz = np.count_nonzero(base)
        with CoreOptions(auto_tidyup=False):
            base = _data.to(datatype, _data.create(base))
        assert np.count_nonzero(base.to_array()) == nnz
        assert not _data.isherm(base, tol=self.tol)
        assert not _data.isherm(base.transpose(), tol=self.tol)

    def test_structure_detection(self, datatype):
        base = np.array([[1,1,0],
                         [0,1,1],
                         [1,0,1]])
        base = _data.to(datatype, _data.create(base))
        assert not _data.isherm(base, tol=self.tol)


class Test_isdiag:
    @pytest.mark.parametrize("shape",
        [(10, 1), (2, 5), (5, 2), (5, 5)]
    )
    def test_isdiag(self, shape, datatype):
        mat = np.zeros(shape)
        data = _data.to(datatype, _data.Dense(mat))
        # empty matrices are diagonal
        assert _data.isdiag(data)

        mat[0, 0] = 1
        data = _data.to(datatype, _data.Dense(mat))
        assert _data.isdiag(data)

        mat[1, 0] = 1
        data = _data.to(datatype, _data.Dense(mat))
        assert not _data.isdiag(data)



#cython: language_level=3

"""
The conversion machinery between different data-layer types, and creation
routines from arbitrary data.  The classes `_to` and `_create` are not intended
to be exported names, but are the backing machinery of the functions `data.to`
and `data.create`, which are built up as the last objects in the `__init__.py`
initialisation of the `data` module.
"""

# This module is compiled by Cython because it's the core of the entire
# dispatch table, and having it compiled to a C extension saves about 1µs per
# call.  This is not much at all, and there's very little which benefits from
# Cython compiliation, but such core functionality is called millions of times
# even in a simple interactive QuTiP session, and it all adds up.

import numbers

import numpy as np
from scipy.sparse import dok_matrix, csgraph
cimport cython
from qutip.core.data.base cimport Data

__all__ = ['to', 'create']


class _Epsilon:
    """
    Constant for an small weight non-null weight.
    Use to set `Data` specialisation just over direct specialisation.
    """
    def __repr__(self):
        return "EPSILON"

    def __eq__(self, other):
        if isinstance(other, _Epsilon):
            return True
        return NotImplemented

    def __add__(self, other):
        if isinstance(other, _Epsilon):
            return self
        return other

    def __radd__(self, other):
        if isinstance(other, _Epsilon):
            return self
        return other

    def __lt__(self, other):
        """ positive number > _Epsilon > 0 """
        if isinstance(other, _Epsilon):
            return False
        return other > 0.

    def __gt__(self, other):
        if isinstance(other, _Epsilon):
            return False
        return other <= 0.


EPSILON = _Epsilon()


def _raise_if_unconnected(dtype_list, weights):
    unconnected = {}
    for i, type_ in enumerate(dtype_list):
        missing = [dtype_list[j].__name__
                   for j, weight in enumerate(weights[:, i])
                   if weight == np.inf]
        if missing:
            unconnected[type_.__name__] = missing
    if unconnected:
        message = "Conversion graph not connected.  Cannot reach:\n * "
        message += "\n * ".join(to + " from (" + ", ".join(froms) + ")"
                                for to, froms in unconnected.items())
        raise NotImplementedError(message)


cdef class _converter:
    """Callable which converts objects of type `x.from_` to type `x.to`."""

    cdef list functions
    cdef Py_ssize_t n_functions
    cdef readonly type to
    cdef readonly type from_

    def __init__(self, functions, to_type, from_type):
        self.functions = list(functions)
        self.n_functions = len(self.functions)
        self.to = to_type
        self.from_ = from_type

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __call__(self, arg):
        if not isinstance(arg, self.from_):
            raise TypeError(str(arg) + " is not of type " + str(self.from_))
        cdef Py_ssize_t i
        for i in range(self.n_functions):
            arg = self.functions[i](arg)
        return arg

    def __repr__(self):
        return ("<converter to "
                + self.to.__name__
                + " from " + self.from_.__name__
                + ">")


def identity_converter(arg):
    return arg


cdef class _partial_converter:
    """Convert from any known data-layer type into the type `x.to`."""

    cdef object converter
    cdef readonly type to

    def __init__(self, converter, to_type):
        self.converter = converter
        self.to = to_type

    def __call__(self, arg):
        try:
            return self.converter[self.to, type(arg)](arg)
        except KeyError:
            raise TypeError("unknown type of input: " + str(arg)) from None

    def __repr__(self):
        return "<converter to " + self.to.__name__ + ">"


# While `_to` and `_create` are defined as objects here, they are actually
# exported by `data.__init__.py` as singleton function objects of their
# respective types (without the leading underscore).

cdef class _to:
    """
    Convert data into a different type.  This object is the knowledge source
    for every allowable data-layer type in QuTiP, and provides the conversions
    between all of them.

    The base use is to call this object as a function with signature
        (type, data) -> converted_data
    where `type` is a type object (such as `data.CSR`, or that obtained by
    calling `type(matrix)`) and `data` is data in a data-layer type.  If you
    want to create a data-layer type from non-data-layer data, use `create`
    instead.

    You can get individual converters by using the key-lookup syntax.  For
    example, the item
        to[CSR, Dense]
    is a callable which accepts arguments of type `Dense` and returns the
    equivalent item of type `CSR`.  You can also get a generic converter to a
    particular data type if only one type is specified, so
        to[Dense]
    is a callable which accepts all known (at the time of the lookup)
    data-layer types, and converts them to `Dense`.  See the `Efficiency notes`
    section below for more detail.

    Internally, the conversion process may go through several steps if new
    data-layer types have been defined with few conversions specified between
    them and the pre-existing converters.  The first-class QuTiP data types
    `Dense` and `CSR` will typically have the best connectivity.


    Adding new types
    ----------------
    You can add new data-layer types by calling the `add_conversions` method of
    this object, and then rebuilding all of the mathematical dispatchers.  See
    the docstring of that method for more information.


    Efficiency notes
    ----------------
    From an efficiency perspective, there is very little benefit to using the
    key-lookup syntax.  Internally, `to(to_type, data)` effectively calls
    `to[to_type, type(data)]`, so storing the object elides the creation of a
    single tuple and a dict lookup, but the cost of this is generally less than
    500ns.  Using the one-argument lookup (e.g. `to[Dense]`) is no more
    efficient than the general call at all, but can be used in cases where a
    single callable is required and is more efficient than `functools.partial`.
    """

    cdef readonly set dtypes
    cdef readonly list dispatchers
    cdef dict _direct_convert
    cdef dict _convert
    cdef readonly dict weight
    cdef readonly dict _str2type

    def __init__(self):
        self._direct_convert = {}
        self._convert = {}
        self.dtypes = set()
        self.weight = {}
        self.dispatchers = []
        self._str2type = {}

    def add_conversions(self, converters):
        """
        Add conversion functions between different data types.  This is an
        advanced function, and is only intended for the QuTiP user who wants to
        add a new underlying data type to QuTiP.

        Any new data type must have at least one converter function given to
        produce the new data type from an existing data type, and at least one
        which produces an existing data type from the new one.  You need not
        specify any more than this, although for efficiency reasons, you may
        want to specify direct conversions for all types you expect the new
        type to interact with frequently.

        Parameters
        ----------
        converters : iterable of (to_type, from_type, converter, [weight])
            An iterable of 3- or 4-tuples describing all the new conversions.
            Each element can individually be a 3- or 4-tuple; they do not need
            to be all one or the other.

            Elements
            ........
            to_type : type
                The data-layer type that is output by the converter.

            from_type : type
                The data-layer type to be input to the converter.

            converter : callable (Data -> Data)
                The converter function.  This should take a single argument
                (the input data-layer function) and output a data-layer object
                of `to_type`.  The converter may assume without error checking
                that its input will always be of `to_type`.  It is safe to
                specify the same conversion function for multiple inputs so
                long as the function handles them all safely, but it must
                always return a single output type.

            weight : positive real, optional (1)
                The weight associated with this conversion.  This must be > 0,
                and defaults to `1` if not supplied (which is fixed to be the
                cost of conversion to `Dense` from `CSR`).  It is generally
                safe just to leave this blank; it is always at best an
                approximation.  The currently defined weights are accessible in
                the `weights` attribute of this object.
                Weight of ~0.001 are should be used in case when no conversion
                is needed or ``converter = lambda mat : mat``.
        """
        for arg in converters:
            if len(arg) == 3:
                to_type, from_type, converter = arg
                weight = 1
            elif len(arg) == 4:
                to_type, from_type, converter, weight = arg
            else:
                raise TypeError("unknown converter specification: " + str(arg))
            if not isinstance(to_type, type):
                raise TypeError(repr(to_type) + " is not a type object")
            if not isinstance(from_type, type):
                raise TypeError(repr(from_type) + " is not a type object")
            if not isinstance(weight, numbers.Real) or weight <= 0:
                raise TypeError("weight " + repr(weight) + " is not valid")
            self.dtypes.add(from_type)
            self.dtypes.add(to_type)
            self._direct_convert[(to_type, from_type)] = (converter, weight)
        # Two-way mapping to convert between the type of a dtype and an integer
        # enumeration value for it.
        order, index = [], {}
        for i, dtype in enumerate(self.dtypes):
            order.append(dtype)
            index[dtype] = i
        # Treat the conversion problem as a shortest-path graph problem.  We
        # build up the graph description as a matrix, then solve the
        # all-pairs-shortest-path problem.  We forbid negative weights and
        # there are unlikely to be many data types, so the choice of algorithm
        # is unimportant (Dijkstra's, Floyd--Warshall, Bellman--Ford, etc).
        graph = dok_matrix((len(order), len(order)))
        for (to_type, from_type), (_, weight) in self._direct_convert.items():
            graph[index[from_type], index[to_type]] = weight
        weights, predecessors =\
            csgraph.floyd_warshall(graph.tocsr(), return_predecessors=True)
        _raise_if_unconnected(order, weights)
        # Build the whole shortest path conversion lookup.  We directly store
        # all complete shortest paths, even though this is not the most memory
        # efficient, because we expect that there will generally be a small
        # number of available data types, and we care more about minimising the
        # number of lookups required.
        self.weight = {}
        self._convert = {}
        for i, from_t in enumerate(order):
            for j, to_t in enumerate(order):
                convert = []
                weight = 0
                cur_t = to_t
                pred_i = predecessors[i, j]
                while pred_i >= 0:
                    pred_t = order[pred_i]
                    _convert, _weight = self._direct_convert[(cur_t, pred_t)]
                    convert.append(_convert)
                    weight += _weight
                    cur_t = pred_t
                    pred_i = predecessors[i, pred_i]
                self.weight[(to_t, from_t)] = weight
                self._convert[(to_t, from_t)] =\
                    _converter(convert[::-1], to_t, from_t)
        for dtype in self.dtypes:
            self.weight[(dtype, Data)] = 1.
            self.weight[(Data, dtype)] = EPSILON
            self._convert[(dtype, Data)] = _partial_converter(self, dtype)
            self._convert[(Data, dtype)] = identity_converter
        for dispatcher in self.dispatchers:
            dispatcher.rebuild_lookup()

    def register_aliases(self, aliases, layer_type):
        """
        Register a user frendly name for a data-layer type to be recognized by
        the :meth:`parse` method.

        Parameters
        ----------
        aliases : str or list of str
            Name of list of names to be understood to represent the layer_type.

        layer_type : type
            Data-layer type, must have been registered with
            :meth:`add_conversions` first.
        """
        if layer_type not in self.dtypes:
            raise ValueError(
                "Type is not a data-layer type: " + repr(layer_type))
        if isinstance(aliases, str):
            aliases = [aliases]
        for alias in aliases:
            if type(alias) is not str:
                raise TypeError("The alias must be a str : " + repr(alias))
            self._str2type[alias] = layer_type

    def parse(self, dtype):
        """
        Return a data-layer type object given its name or the type itself.

        Parameters
        ----------
        dtype : type, str
            Either the name of a data-layer type or a type itself.

        Returns
        -------
        type
            A data-layer type.

        Raises
        ------
        TypeError
            If ``dtype`` is neither a string nor a type.
        ValueError
            If ``dtype`` is a name, but no data-layer type of that name is
            registered, or if ``dtype`` is a type, but not a known data-layer
            type.
        """
        if type(dtype) is type:
            if dtype not in self.dtypes and dtype is not Data:
                raise ValueError(
                    "Type is not a data-layer type: " + repr(dtype))
            return dtype
        elif type(dtype) is str:
            try:
                return self._str2type[dtype]
            except KeyError:
                raise ValueError(
                    "Type name is not known to the data-layer: " + repr(dtype)
                    ) from None

        raise TypeError(
            "Invalid dtype is neither a type nor a type name: " + repr(dtype))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __getitem__(self, arg):
        if type(arg) is not tuple:
            arg = (arg,)
        if not arg or len(arg) > 2:
            raise KeyError(arg)
        to_t = self.parse(arg[0])
        if len(arg) == 1:
            return _partial_converter(self, to_t)
        from_t = self.parse(arg[1])
        return self._convert[to_t, from_t]

    def __call__(self, to_type, data):
        to_type = self.parse(to_type)
        from_type = self.parse(type(data))
        if to_type == from_type:
            return data
        return self._convert[to_type, from_type](data)


cdef class _create:
    cdef readonly list _creators

    def __init__(self):
        self._creators = []

    def add_creators(self, creators):
        """
        Add creation functions to make a data-layer object from an arbitrary
        Python object.

        Parameters
        ----------
        creators : iterable of (condition, creator, [priority])
            An iterable of 2- or 3-tuples describing the new data layer
            creation functions.
            Each element can individually be a 2- or 3-tuple; they do not need
            to be all one or the other.

            Elements
            ........
            condition : callable (object) -> bool
                Function determining if the given object can be converted to a
                data-layer type using this creator.

            creator function: callable (object, shape, copy=True) -> Data
                The creator function. It should take an object and a shape and
                return a data-layer type instance. The object may be any object
                for which the condition function returned ``True`` when tested.
                The ``object`` and ``shape`` parameters are passed positionally,
                and the ``copy`` parameter is passed by keyword.

            priority : real, optional (0)
                The priority associated with this creator. Higher priority
                conditions will be tested first and the first valid creator
                (i.e. for which ``condition(object) == True``) will handle
                the creation.

        Notes
        -----
        Default creators are added with the following priorities:

        * Objects that are instances of data-layer types are
          converted using ``.copy`` with priority 100.
        * Objects that have a direct equivalent such as ``numpy.ndarray``
          or ``scipy.sparse.csr_matrix`` are converted with priority 80.
        * Objects for which ``scipy.sparse.issparse`` is ``True``
          are converted using an internal CSR converter with
          priority 20.
        * If no condition are meet, ``numpy.array`` is used to try convert
          the input to an array (priority -inf).
        """
        for condition, creator, *priority in creators:
            if not callable(condition):
                raise TypeError(repr(condition) + " is not a callable")
            if not callable(create):
                raise TypeError(repr(create) + " is not a callable")
            if len(priority) >= 2:
                raise ValueError("Too many values to unpack for a creator, " +
                                 "expected 2 or 3, got "+ str(2+len(priority)))
            priority = float(priority[0] if priority else 0)
            self._creators.append((condition, creator, priority))
        self._creators.sort(key=lambda creator: creator[2], reverse=True)

    def __call__(self, arg, shape=None, copy=True):
        """
        Build a :class:`.Data` object from arg.

        Parameters
        ----------
        arg : object
            Object to be converted to `qutip.data.Data`. Anything that can be
            converted to a numpy array are valid input.

        shape : tuple, optional
            The shape of the output as (``rows``, ``columns``).
        """
        for condition, create, _ in self._creators:
            if condition(arg):
                return create(arg, shape, copy=copy)
        raise TypeError(f"arg `{repr(arg)}` cannot be converted to "
                        "qutip data format")


to = _to()
create = _create()



__all__ = ['krylovsolve']

from .. import QobjEvo
from .sesolve import SESolver


def krylovsolve(
    H, psi0, tlist, krylov_dim, e_ops=None, args=None, options=None
):
    """
    Schrodinger equation evolution of a state vector for time independent
    Hamiltonians using Krylov method.

    Evolve the state vector ("psi0") finding an approximation for the time
    evolution operator of Hamiltonian ("H") by obtaining the projection of
    the time evolution operator on a set of small dimensional Krylov
    subspaces (m << dim(H)).

    The output is either the state vector or unitary matrix at arbitrary points
    in time (`tlist`), or the expectation values of the supplied operators
    (`e_ops`). If e_ops is a callback function, it is invoked for each
    time in `tlist` with time and the state as arguments, and the function
    does not use any return values. e_ops cannot be used in conjunction
    with solving the Schrodinger operator equation

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        System Hamiltonian as a Qobj or QobjEvo for time-dependent
        Hamiltonians. List of [:obj:`.Qobj`, :obj:`.Coefficient`] or callable
        that can be made into :obj:`.QobjEvo` are also accepted.

    psi0 : :class:`.Qobj`
        Initial state vector (ket)

    tlist : *list* / *array*
        list of times for :math:`t`.

    krylov_dim: int
        Dimension of Krylov approximation subspaces used for the time
        evolution approximation.

    e_ops : :class:`.Qobj`, callable, or list, optional
        Single operator or list of operators for which to evaluate
        expectation values or callable or list of callable.
        Callable signature must be, `f(t: float, state: Qobj)`.
        See :func:`~qutip.core.expect.expect` for more detail of operator
        expectation.

    args : dict, optional
        dictionary of parameters for time-dependent Hamiltonians

    options : dict, optional
        Dictionary of options for the solver.

        - | store_final_state : bool
          | Whether or not to store the final state of the evolution in the
            result class.
        - | store_states : bool, None
          | Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.
        - | normalize_output : bool
          | Normalize output state to hide ODE numerical errors.
        - | progress_bar : str {'text', 'enhanced', 'tqdm', ''}
          | How to present the solver progress.
            'tqdm' uses the python module of the same name and raise an error
            if not installed. Empty string or False will disable the bar.
        - | progress_kwargs : dict
          | kwargs to pass to the progress_bar. Qutip's bars use `chunk_size`.
        - | atol: float
          | Absolute tolerance of the ODE integrator.
        - | nsteps : int
          | Maximum number of (internally defined) steps allowed in one
            ``tlist`` step.
        - | min_step, max_step : float
          | Miniumum and maximum lenght of one internal step.
        - | always_compute_step: bool
          | If True, the step lenght is computed each time a new Krylov
            subspace is computed. Otherwise it is computed only once when
            creating the integrator.
        - | sub_system_tol: float
          | Tolerance to detect an happy breakdown. An happy breakdown happens
            when the initial ket is in a subspace of the Hamiltonian smaller
            than ``krylov_dim``.

    Returns
    -------
    result: :class:`.Result`

        An instance of the class :class:`.Result`, which contains
        a *list of array* ``result.expect`` of expectation values for the times
        specified by ``tlist``, and/or a *list* ``result.states`` of state
        vectors or density matrices corresponding to the times in ``tlist`` [if
        ``e_ops`` is an empty list of ``store_states=True`` in options].
    """
    H = QobjEvo(H, args=args, tlist=tlist)
    options = options or {}
    options["method"] = "krylov"
    options["krylov_dim"] = krylov_dim
    solver = SESolver(H, options=options)
    return solver.run(psi0, tlist, e_ops=e_ops)



import numpy as np
from types import FunctionType

import qutip
from qutip.solver.mesolve import mesolve, MESolver
from qutip.solver.solver_base import Solver
import pickle
import pytest

# Deactivate warning for test without cython
from qutip.core.coefficient import WARN_MISSING_MODULE
WARN_MISSING_MODULE[0] = 0


all_ode_method = [
    method for method, integrator in MESolver.avail_integrators().items()
    if integrator.support_time_dependant
]

def fidelitycheck(out1, out2, rho0vec):
    fid = np.zeros(len(out1.states))
    for i, E in enumerate(out2.states):
        rhot = qutip.vector_to_operator(E*rho0vec)
        fid[i] = qutip.fidelity(out1.states[i], rhot)
    return fid


class TestMESolveDecay:
    N = 10
    a = qutip.destroy(N)
    kappa = 0.2
    tlist = np.linspace(0, 10, 201)
    ada = a.dag() * a

    @pytest.fixture(params=[
        pytest.param([ada, lambda t, args: 1], id='Hlist_func'),
        pytest.param([ada, '1'], id='Hlist_str'),
        pytest.param([ada, np.ones_like(tlist)], id='Hlist_array'),
        pytest.param(qutip.QobjEvo([ada, '1']), id='HQobjEvo'),
        pytest.param(lambda t, args: qutip.create(10) * qutip.destroy(10),
                     id='func'),
    ])
    def H(self, request):
        return request.param

    @pytest.fixture(params=[
        pytest.param(np.sqrt(kappa) * a,
                     id='const'),
        pytest.param(lambda t, args: (np.sqrt(args['kappa'])
                                      * qutip.destroy(10)),
                     id='func'),
        pytest.param([a, lambda t, args: np.sqrt(args['kappa'])],
                     id='list_func'),
        pytest.param([a, 'sqrt(kappa)'],
                     id='list_str'),
        pytest.param([a, np.sqrt(kappa) * np.ones_like(tlist)],
                     id='list_array'),
        pytest.param(qutip.QobjEvo([a, 'sqrt(kappa)'], args={'kappa': kappa}),
                     id='QobjEvo'),
    ])
    def cte_c_ops(self, request):
        return request.param

    @pytest.fixture(params=[
        pytest.param([a, lambda t, args: np.sqrt(args['kappa'] * np.exp(-t))],
                  id='list_func'),
        pytest.param([a, 'sqrt(kappa * exp(-t))'],
                  id='list_str'),
        pytest.param([a, np.sqrt(kappa * np.exp(-tlist))],
                  id='list_array'),
        pytest.param(qutip.QobjEvo([a, 'sqrt(kappa * exp(-t))'],
                          args={'kappa': kappa}),
                  id='QobjEvo'),
        pytest.param(lambda t, args: (np.sqrt(args['kappa'] * np.exp(-t)) *
                                      qutip.destroy(10)),
                     id='func'),
    ])
    def c_ops(self, request):
        return request.param

    c_ops_1 = c_ops

    def testME_CteDecay(self, cte_c_ops):
        "mesolve: simple constant decay"
        me_error = 1e-6
        H = self.a.dag() * self.a
        psi0 = qutip.basis(self.N, 9)  # initial state
        c_op_list = [cte_c_ops]
        options = {"progress_bar": None}
        medata = mesolve(H, psi0, self.tlist, c_op_list, [H],
                         args={"kappa": self.kappa},
                         options=options)
        expt = medata.expect[0]
        actual_answer = 9.0 * np.exp(-self.kappa * self.tlist)
        np.testing.assert_allclose(actual_answer, expt, atol=me_error)

    @pytest.mark.parametrize('method',
                             all_ode_method, ids=all_ode_method)
    def testME_TDDecay(self, c_ops, method):
        "mesolve: time-dependence as function list"
        me_error = 1e-5
        H = self.a.dag() * self.a
        psi0 = qutip.basis(self.N, 9)  # initial state
        c_op_list = [c_ops]
        options = {"method": method, "progress_bar": None}
        medata = mesolve(H, psi0, self.tlist, c_op_list, [H],
                         args={"kappa": self.kappa}, options=options)
        expt = medata.expect[0]
        actual_answer = 9.0 * np.exp(-self.kappa * (1.0 - np.exp(-self.tlist)))
        np.testing.assert_allclose(actual_answer, expt, rtol=me_error)

    def testME_2TDDecay(self, c_ops, c_ops_1):
        "mesolve: time-dependence as function list"
        me_error = 1e-5
        H = self.a.dag() * self.a
        psi0 = qutip.basis(self.N, 9)  # initial state
        c_op_list = [c_ops, c_ops_1]
        options = {"progress_bar": None}
        medata = mesolve(H, psi0, self.tlist, c_op_list, [H],
                         args={"kappa": self.kappa},
                         options=options)
        expt = medata.expect[0]
        actual_answer = 9.0 * np.exp(-2 * self.kappa *
                                     (1.0 - np.exp(-self.tlist)))
        np.testing.assert_allclose(actual_answer, expt, atol=me_error)

    def testME_TDH_TDDecay(self, H, c_ops):
        "mesolve: time-dependence as function list"
        me_error = 5e-6
        psi0 = qutip.basis(self.N, 9)  # initial state
        c_op_list = [c_ops]
        options = {"progress_bar": None}
        medata = mesolve(H, psi0, self.tlist, c_op_list, [self.ada],
                         args={"kappa": self.kappa},
                         options=options)
        expt = medata.expect[0]
        actual_answer = 9.0 * np.exp(-self.kappa *
                                     (1.0 - np.exp(-self.tlist)))
        np.testing.assert_allclose(actual_answer, expt, atol=me_error)

    def testME_TDH_longTDDecay(self, H, c_ops):
        "mesolve: time-dependence as function list"
        me_error = 2e-5
        psi0 = qutip.basis(self.N, 9)  # initial state
        if isinstance(c_ops, FunctionType):
            return
        if isinstance(c_ops, qutip.QobjEvo):
            c_op_list = [c_ops + c_ops]
        else:
            c_op_list = [[c_ops, c_ops]]
        options = {"progress_bar": None}
        medata = mesolve(H, psi0, self.tlist, c_op_list, [self.ada],
                         args={"kappa": self.kappa},
                         options=options)
        expt = medata.expect[0]
        actual_answer = 9.0 * np.exp(-4 * self.kappa *
                                     (1.0 - np.exp(-self.tlist)))
        np.testing.assert_allclose(actual_answer, expt, atol=me_error)

    def testME_TDDecayUnitary(self, c_ops):
        "mesolve: time-dependence as function list with super as init cond"
        me_error = 5e-6

        H = self.ada
        psi0 = qutip.basis(self.N, 9)  # initial state
        rho0vec = qutip.operator_to_vector(psi0*psi0.dag())
        E0 = qutip.sprepost(qutip.qeye(self.N), qutip.qeye(self.N))
        options = {"progress_bar": None}
        c_op_list = [c_ops]
        out1 = mesolve(H, psi0, self.tlist, c_op_list, [],
                       args={"kappa": self.kappa},
                       options=options)
        out2 = mesolve(H, E0, self.tlist, c_op_list, [],
                       args={"kappa": self.kappa},
                       options=options)

        fid = fidelitycheck(out1, out2, rho0vec)
        assert fid == pytest.approx(1., abs=me_error)

    def testME_TDDecayliouvillian(self, c_ops):
        "mesolve: time-dependence as function list with super as init cond"
        me_error = 5e-6

        H = self.ada
        L = qutip.liouvillian(H)
        psi0 = qutip.basis(self.N, 9)  # initial state
        rho0vec = qutip.operator_to_vector(psi0*psi0.dag())
        E0 = qutip.sprepost(qutip.qeye(self.N), qutip.qeye(self.N))
        options = {"progress_bar": None}
        c_op_list = [c_ops]
        out1 = mesolve(L, psi0, self.tlist, c_op_list, [],
                       args={"kappa": self.kappa},
                       options=options)
        out2 = mesolve(L, E0, self.tlist, c_op_list, [],
                       args={"kappa": self.kappa},
                       options=options)

        fid = fidelitycheck(out1, out2, rho0vec)
        assert fid == pytest.approx(1., abs=me_error)

    @pytest.mark.parametrize(['state_type'], [
        pytest.param("ket", id="ket"),
        pytest.param("dm", id="dm"),
        pytest.param("unitary", id="unitary"),
    ])
    def test_mesolve_normalization(self, state_type):
        # non-hermitean H causes state to evolve non-unitarily
        H = qutip.Qobj([[1, -0.1j], [-0.1j, 1]])
        H = qutip.sprepost(H, H) # ensure use of MeSolve
        psi0 = qutip.basis(2, 0)
        options = {"normalize_output": True, "progress_bar": None}

        if state_type in {"ket", "dm"}:
            if state_type == "dm":
                psi0 = qutip.ket2dm(psi0)
            output = mesolve(H, psi0, self.tlist, e_ops=[], options=options)
            norms = [state.norm() for state in output.states]
            np.testing.assert_allclose(
                norms, [1.0 for _ in self.tlist], atol=1e-15,
            )
        else:
            # evolution of unitaries should not be normalized
            U = qutip.sprepost(qutip.qeye(2), qutip.qeye(2))
            output = mesolve(H, U, self.tlist, e_ops=[], options=options)
            norms = [state.norm() for state in output.states]
            assert all(norm > 4 for norm in norms[1:])

    def test_mesolver_pickling(self):
        options = {"progress_bar": None}
        solver_obj = MESolver(self.ada, c_ops=[self.a], options=options)
        solver_copy = pickle.loads(pickle.dumps(solver_obj))
        e1 = solver_obj.run(qutip.basis(self.N, 9), [0, 1, 2, 3],
                            e_ops=[self.ada]).expect[0]
        e2 = solver_copy.run(qutip.basis(self.N, 9), [0, 1, 2, 3],
                            e_ops=[self.ada]).expect[0]
        np.testing.assert_allclose(e1, e2)

    @pytest.mark.parametrize('method',
                             all_ode_method, ids=all_ode_method)
    def test_mesolver_stepping(self, method):
        options = {"method": method, "progress_bar": None}
        solver_obj = MESolver(
            self.ada,
            c_ops=qutip.QobjEvo(
                [self.a, lambda t, kappa: np.sqrt(kappa * np.exp(-t))],
                args={'kappa': 1.}
            ),
            options=options
        )
        solver_obj.start(qutip.basis(self.N, 9), 0)
        state1 = solver_obj.step(1)
        assert qutip.expect(self.ada, state1) != (
            qutip.expect(self.ada, qutip.basis(self.N, 9))
        )
        state2 = solver_obj.step(2, args={"kappa":0.})
        np.testing.assert_allclose(qutip.expect(self.ada, state1),
                                   qutip.expect(self.ada, state2), 1e-6)


@pytest.mark.parametrize('super_', ["ket", "dm", "liouvillian"],
                         ids=["ket", "dm", "liouvillian"])
def testME_SesolveFallback(super_):
    "mesolve: final_state has correct dims"
    N = 5
    a = qutip.tensor(qutip.destroy(N+1), qutip.qeye(N+1), qutip.qeye(N+1))
    b = qutip.tensor(qutip.qeye(N+1), qutip.destroy(N+1), qutip.qeye(N+1))
    c = qutip.tensor(qutip.qeye(N+1), qutip.qeye(N+1), qutip.destroy(N+1))
    psi0 = qutip.tensor(qutip.basis(N+1,0),
                        qutip.basis(N+1,0),
                        qutip.basis(N+1,N))
    H = a * b * c.dag() * c.dag() + a.dag() * b.dag() * c * c
    if super_ == "dm":
        state0 = qutip.ket2dm(psi0)
    else:
        state0 = psi0
    if super_ == "liouvillian":
        H = qutip.liouvillian(H)

    times = np.linspace(0.0, 0.1, 3)
    options = {
        "store_states": False,
        "store_final_state": True,
        "progress_bar": None
    }
    result = mesolve(H, state0, times, [], e_ops=[a], options=options)
    if super_ == "ket":
        assert result.final_state.dims == psi0.dims
    else:
        assert result.final_state.dims == qutip.ket2dm(psi0).dims


class TestJCModelEvolution:
    """
    A test class for the QuTiP functions for the evolution of JC model
    """
    def qubit_integrate(self, tlist, psi0, epsilon, delta, g1, g2):

        H = epsilon / 2.0 * qutip.sigmaz() + delta / 2.0 * qutip.sigmax()

        c_op_list = []

        rate = g1
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * qutip.sigmam())

        rate = g2
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * qutip.sigmaz())

        output = mesolve(H, psi0, tlist, c_op_list,
            e_ops=[qutip.sigmax(), qutip.sigmay(), qutip.sigmaz()]
        )

        return output.expect[0], output.expect[1], output.expect[2]

    def jc_steadystate(self, N, wc, wa, g, kappa, gamma,
                       pump, psi0, use_rwa, tlist):

        # Hamiltonian
        a = qutip.tensor(qutip.destroy(N), qutip.identity(2))
        sm = qutip.tensor(qutip.identity(N), qutip.destroy(2))

        if use_rwa:
            # use the rotating wave approxiation
            H = (wc * a.dag() * a + wa * sm.dag() * sm
                 + g * (a.dag() * sm + a * sm.dag()))
        else:
            H = (wc * a.dag() * a + wa * sm.dag() * sm
                 + g * (a.dag() + a) * (sm + sm.dag()))

        # collapse operators
        c_op_list = []

        n_th_a = 0.0  # zero temperature

        rate = kappa * (1 + n_th_a)
        c_op_list.append(np.sqrt(rate) * a)

        rate = kappa * n_th_a
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * a.dag())

        rate = gamma
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * sm)

        rate = pump
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * sm.dag())

        # find the steady state
        rho_ss = qutip.steadystate(H, c_op_list)

        return (qutip.expect(a.dag() * a, rho_ss),
                qutip.expect(sm.dag() * sm, rho_ss))

    def jc_integrate(self, N, wc, wa, g, kappa, gamma,
                     pump, psi0, use_rwa, tlist, oper_evo=False):

        # Hamiltonian
        a = qutip.tensor(qutip.destroy(N), qutip.identity(2))
        sm = qutip.tensor(qutip.identity(N), qutip.destroy(2))

        # Identity super-operator
        E0 = qutip.sprepost(
            qutip.tensor(qutip.qeye(N), qutip.qeye(2)),
            qutip.tensor(qutip.qeye(N), qutip.qeye(2))
        )

        if use_rwa:
            # use the rotating wave approxiation
            H = wc * a.dag() * a + wa * sm.dag() * sm + g * (
                a.dag() * sm + a * sm.dag())
        else:
            H = wc * a.dag() * a + wa * sm.dag() * sm + g * (
                a.dag() + a) * (sm + sm.dag())

        # collapse operators
        c_op_list = []

        n_th_a = 0.0  # zero temperature

        rate = kappa * (1 + n_th_a)
        c_op_list.append(np.sqrt(rate) * a)

        rate = kappa * n_th_a
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * a.dag())

        rate = gamma
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * sm)

        rate = pump
        if rate > 0.0:
            c_op_list.append(np.sqrt(rate) * sm.dag())

        options = {"store_states": True, "progress_bar": None}

        # evolve and calculate expectation values
        output = mesolve(
            H, psi0, tlist, c_op_list, [a.dag() * a, sm.dag() * sm],
            options=options)
        if oper_evo:
            output2 = mesolve(H, E0, tlist, c_op_list, [])
            return output, output2
        return output.expect[0], output.expect[1]

    def testSuperJC(self):
        "mesolve: super vs. density matrix as initial condition"
        me_error = 1e-6

        use_rwa = True
        N = 4           # number of cavity fock states
        wc = 2 * np.pi * 1.0   # cavity frequency
        wa = 2 * np.pi * 1.0   # atom frequency
        g = 2 * np.pi * 0.1   # coupling strength
        kappa = 0.05    # cavity dissipation rate
        gamma = 0.001   # atom dissipation rate
        pump = 0.25    # atom pump rate

        # start with an excited atom and maximum number of photons
        n = N - 2
        psi0 = qutip.basis([N, 2], [n, 1])
        rho0vec = qutip.operator_to_vector(psi0.proj())
        tlist = np.linspace(0, 100, 50)

        out1, out2 = self.jc_integrate(
            N, wc, wa, g, kappa, gamma, pump, psi0, use_rwa, tlist, True)

        fid = fidelitycheck(out1, out2, rho0vec)
        assert fid == pytest.approx(1., abs=me_error)

    def testQubitDynamics1(self):
        "mesolve: qubit with dissipation"

        epsilon = 0.0 * 2 * np.pi   # cavity frequency
        delta = 1.0 * 2 * np.pi   # atom frequency
        g2 = 0.1
        g1 = 0.0
        psi0 = qutip.basis(2, 0)        # initial state
        tlist = np.linspace(0, 5, 200)

        sx, sy, sz = self.qubit_integrate(tlist, psi0, epsilon, delta, g1, g2)

        sx_analytic = np.zeros(np.shape(tlist))
        sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
        sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

        np.testing.assert_allclose(sx, sx_analytic, atol=0.05)
        np.testing.assert_allclose(sy, sy_analytic, atol=0.05)
        np.testing.assert_allclose(sz, sz_analytic, atol=0.05)

    def testQubitDynamics2(self):
        "mesolve: qubit without dissipation"

        epsilon = 0.0 * 2 * np.pi   # cavity frequency
        delta = 1.0 * 2 * np.pi   # atom frequency
        g2 = 0.0
        g1 = 0.0
        psi0 = qutip.basis(2, 0)        # initial state
        tlist = np.linspace(0, 5, 200)

        sx, sy, sz = self.qubit_integrate(tlist, psi0, epsilon, delta, g1, g2)

        sx_analytic = np.zeros(np.shape(tlist))
        sy_analytic = -np.sin(2 * np.pi * tlist) * np.exp(-tlist * g2)
        sz_analytic = np.cos(2 * np.pi * tlist) * np.exp(-tlist * g2)

        np.testing.assert_allclose(sx, sx_analytic, atol=0.05)
        np.testing.assert_allclose(sy, sy_analytic, atol=0.05)
        np.testing.assert_allclose(sz, sz_analytic, atol=0.05)

    def testCavity1(self):
        "mesolve: cavity-qubit interaction, no dissipation"

        use_rwa = True
        N = 4           # number of cavity fock states
        wc = 2 * np.pi * 1.0   # cavity frequency
        wa = 2 * np.pi * 1.0   # atom frequency
        g = 2 * np.pi * 0.01   # coupling strength
        kappa = 0.0     # cavity dissipation rate
        gamma = 0.0     # atom dissipation rate
        pump = 0.0      # atom pump rate

        # start with an excited atom and maximum number of photons
        n = N - 2
        psi0 = qutip.tensor(qutip.basis(N, n), qutip.basis(2, 1))
        tlist = np.linspace(0, 1000, 2000)

        nc, na = self.jc_integrate(
            N, wc, wa, g, kappa, gamma, pump, psi0, use_rwa, tlist)

        nc_ex = 0.5 * (1 - np.cos(2 * g * np.sqrt(n + 1) * tlist)) + n
        na_ex = 0.5 * (1 + np.cos(2 * g * np.sqrt(n + 1) * tlist))

        np.testing.assert_allclose(nc[-1], nc_ex[-1], atol=0.005)
        np.testing.assert_allclose(na[-1], na_ex[-1], atol=0.005)

    def testCavity2(self):
        "mesolve: cavity-qubit without interaction, decay"

        use_rwa = True
        N = 4           # number of cavity fock states
        wc = 2 * np.pi * 1.0   # cavity frequency
        wa = 2 * np.pi * 1.0   # atom frequency
        g = 2 * np.pi * 0.0    # coupling strength
        kappa = 0.005   # cavity dissipation rate
        gamma = 0.01    # atom dissipation rate
        pump = 0.0      # atom pump rate

        # start with an excited atom and maximum number of photons
        n = N - 2
        psi0 = qutip.tensor(qutip.basis(N, n), qutip.basis(2, 1))
        tlist = np.linspace(0, 1000, 2000)

        nc, na = self.jc_integrate(
            N, wc, wa, g, kappa, gamma, pump, psi0, use_rwa, tlist)

        nc_ex = (0.5 * (1 - np.cos(2 * g * np.sqrt(n + 1) * tlist)) + n) * \
            np.exp(-kappa * tlist)
        na_ex = 0.5 * (1 + np.cos(2 * g * np.sqrt(n + 1) * tlist)) * \
            np.exp(-gamma * tlist)

        np.testing.assert_allclose(nc[-1], nc_ex[-1], atol=0.005)
        np.testing.assert_allclose(na[-1], na_ex[-1], atol=0.005)

    def testCavity3(self):
        "mesolve: cavity-qubit with interaction, decay"

        use_rwa = True
        N = 4           # number of cavity fock states
        wc = 2 * np.pi * 1.0   # cavity frequency
        wa = 2 * np.pi * 1.0   # atom frequency
        g = 2 * np.pi * 0.1   # coupling strength
        kappa = 0.05    # cavity dissipation rate
        gamma = 0.001   # atom dissipation rate
        pump = 0.25    # atom pump rate

        # start with an excited atom and maximum number of photons
        n = N - 2
        psi0 = qutip.tensor(qutip.basis(N, n), qutip.basis(2, 1))
        tlist = np.linspace(0, 200, 500)

        nc, na = self.jc_integrate(
            N, wc, wa, g, kappa, gamma, pump, psi0, use_rwa, tlist)

        # we don't have any analytics for this parameters, so
        # compare with the steady state
        nc_ss, na_ss = self.jc_steadystate(
            N, wc, wa, g, kappa, gamma, pump, psi0, use_rwa, tlist)

        nc_ss = nc_ss * np.ones(np.shape(nc))
        na_ss = na_ss * np.ones(np.shape(na))

        np.testing.assert_allclose(nc[-1], nc_ss[-1], atol=0.005)
        np.testing.assert_allclose(na[-1], na_ss[-1], atol=0.005)


class TestMESolveStepFuncCoeff:
    """
    A Test class for using time-dependent array coefficients
    as step functions instead of doing interpolation
    """
    # Runge-Kutta method (dop853) behave better with step function evolution
    # than multi-step methods (adams, qutip 4's default)
    options = {"method": "dop853", "nsteps": 1e8, "progress_bar": None}

    def python_coeff(self, t, args):
        if t < np.pi/2:
            return 1.
        else:
            return 0.

    @pytest.mark.parametrize('method',
                             all_ode_method, ids=all_ode_method)
    def test_py_coeff(self, method):
        """
        Test for Python function as coefficient as step function coeff
        """
        rho0 = qutip.rand_ket(2)
        tlist = np.array([0, np.pi/2])
        options = {"method": method, "nsteps": 1e5, "rtol": 1e-7}
        qu = qutip.QobjEvo([[qutip.sigmax(), self.python_coeff]],
                     tlist=tlist, order=0)
        result = mesolve(qu, rho0=rho0, tlist=tlist, options=options)
        fid = qutip.fidelity(result.states[-1], qutip.sigmax()*rho0)
        assert fid == pytest.approx(1)

    def test_array_cte_coeff(self):
        """
        Test for Array coefficient with uniform tlist as step function coeff
        """
        rho0 = qutip.rand_ket(2)
        tlist = np.array([0., np.pi/2, np.pi], dtype=float)
        npcoeff = np.array([0.25, 0.75, 0.75])
        qu = qutip.QobjEvo([[qutip.sigmax(), npcoeff]], tlist=tlist, order=0)
        result = mesolve(qu, rho0=rho0, tlist=tlist, options=self.options)
        fid = qutip.fidelity(result.states[-1], qutip.sigmax()*rho0)
        assert fid == pytest.approx(1)

    def test_array_t_coeff(self):
        """
        Test for Array with non-uniform tlist as step function coeff
        """
        rho0 = qutip.rand_ket(2)
        tlist = np.array([0., np.pi/2, np.pi*3/2], dtype=float)
        npcoeff = np.array([0.5, 0.25, 0.25])
        qu = qutip.QobjEvo([[qutip.sigmax(), npcoeff]], tlist=tlist, order=0)
        result = mesolve(qu, rho0=rho0, tlist=tlist, options=self.options)
        fid = qutip.fidelity(result.states[-1], qutip.sigmax()*rho0)
        assert fid == pytest.approx(1)

    def test_array_str_coeff(self):
        """
        Test for Array and string as step function coeff.
        qobjevo_codegen is used and uniform tlist
        """
        rho0 = qutip.rand_ket(2)
        tlist = np.array([0., np.pi/2, np.pi], dtype=float)
        npcoeff1 = np.array([0.25, 0.75, 0.75], dtype=complex)
        npcoeff2 = np.array([0.5, 1.5, 1.5], dtype=float)
        strcoeff = "1."
        qu = qutip.QobjEvo(
            [[qutip.sigmax(), npcoeff1],
             [qutip.sigmax(), strcoeff],
             [qutip.sigmax(), npcoeff2]],
            tlist=tlist, order=0)
        result = mesolve(qu, rho0=rho0, tlist=tlist, options=self.options)
        fid = qutip.fidelity(result.states[-1], qutip.sigmax()*rho0)
        assert fid == pytest.approx(1)

    def test_array_str_py_coeff(self):
        """
        Test for Array, string and Python function as step function coeff.
        qobjevo_codegen is used and non non-uniform tlist
        """
        rho0 = qutip.rand_ket(2)
        tlist = np.array([0., np.pi/4, np.pi/2, np.pi], dtype=float)
        npcoeff1 = np.array([0.4, 1.6, 1.0, 1.0], dtype=complex)
        npcoeff2 = np.array([0.4, 1.6, 1.0, 1.0], dtype=float)
        strcoeff = "1."
        qu = qutip.QobjEvo(
            [[qutip.sigmax(), npcoeff1], [qutip.sigmax(), npcoeff2],
             [qutip.sigmax(), self.python_coeff], [qutip.sigmax(), strcoeff]],
            tlist=tlist, order=0)
        result = mesolve(qu, rho0=rho0, tlist=tlist, options=self.options)
        fid = qutip.fidelity(result.states[-1], qutip.sigmax()*rho0)
        assert fid == pytest.approx(1)


def test_num_collapse_set():
    H = qutip.sigmaz()
    psi = (qutip.basis(2, 0) + qutip.basis(2, 1)).unit()
    ts = [0, 1]
    for c_ops in (
        qutip.sigmax(),
        [qutip.sigmax()],
        [qutip.sigmay(), qutip.sigmax()],
    ):
        res = mesolve(H, psi, ts, c_ops=c_ops)
        if not isinstance(c_ops, list):
            c_ops = [c_ops]
        assert res.stats["num_collapse"] == len(c_ops)


def test_mesolve_bad_H():
    with pytest.raises(TypeError):
        MESolver([qutip.qeye(3), 't'], [])
    with pytest.raises(TypeError):
        MESolver(qutip.qeye(3), [[qutip.qeye(3), 't']])


def test_mesolve_bad_state():
    solver = MESolver(qutip.qeye(4), [])
    with pytest.raises(TypeError):
        solver.start(qutip.basis(2,1) & qutip.basis(2,0), 0)


def test_mesolve_bad_options():
    with pytest.raises(TypeError):
        MESolver(qutip.qeye(4), [], options=False)


def test_feedback():

    def f(t, A):
        return (A-4.)

    N = 10
    tol = 1e-14
    psi0 = qutip.basis(N, 7)
    a = qutip.QobjEvo(
        [qutip.destroy(N), f],
        args={"A": MESolver.ExpectFeedback(qutip.spre(qutip.num(N)))}
    )
    H = qutip.QobjEvo(qutip.num(N))
    solver = qutip.MESolver(H, c_ops=[a])
    result = solver.run(psi0, np.linspace(0, 30, 301), e_ops=[qutip.num(N)])
    assert np.all(result.expect[0] > 4. - tol)
